from __future__ import annotations

import dataclasses
import json

import pytest
from chik_rs import AugSchemeMPL, G1Element, G2Element

from chik._tests.conftest import ConsensusMode
from chik._tests.environments.wallet import WalletStateTransition, WalletTestFramework
from chik._tests.util.setup_nodes import OldSimulatorsAndWallets
from chik._tests.util.time_out_assert import time_out_assert
from chik.rpc.wallet_request_types import DIDGetCurrentCoinInfo, DIDGetRecoveryInfo
from chik.rpc.wallet_rpc_api import WalletRpcApi
from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.simulator.simulator_protocol import FarmNewBlockProtocol
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.condition_opcodes import ConditionOpcode
from chik.types.peer_info import PeerInfo
from chik.types.signing_mode import CHIP_0002_SIGN_MESSAGE_PREFIX
from chik.util.bech32m import decode_puzzle_hash, encode_puzzle_hash
from chik.util.condition_tools import conditions_dict_for_solution
from chik.util.ints import uint16, uint32, uint64
from chik.wallet.did_wallet.did_wallet import DIDWallet
from chik.wallet.singleton import create_singleton_puzzle
from chik.wallet.util.address_type import AddressType
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG
from chik.wallet.util.wallet_types import WalletType
from chik.wallet.wallet_node import WalletNode
from chik.wallet.wallet_spend_bundle import WalletSpendBundle


async def get_wallet_num(wallet_manager):
    return len(await wallet_manager.get_all_wallet_info_entries())


def get_parent_num(did_wallet: DIDWallet):
    return len(did_wallet.did_info.parent_info)


class TestDIDWallet:
    #  TODO: See Issue CHIK-1544
    #  This test should be ported to WalletTestFramework once we can replace keys in the wallet node
    @pytest.mark.parametrize(
        "trusted",
        [True, False],
    )
    @pytest.mark.anyio
    async def test_creation_from_coin_spend(
        self, self_hostname: str, two_nodes_two_wallets_with_same_keys: OldSimulatorsAndWallets, trusted: bool
    ) -> None:
        """
        Verify that DIDWallet.create_new_did_wallet_from_coin_spend() is called after Singleton creation on
        the blockchain, and that the wallet is created in the second wallet node.
        """
        full_nodes, wallets, _ = two_nodes_two_wallets_with_same_keys
        full_node_api = full_nodes[0]
        full_node_server = full_node_api.server
        wallet_node_0, server_0 = wallets[0]
        wallet_node_1, server_1 = wallets[1]

        wallet_0 = wallet_node_0.wallet_state_manager.main_wallet
        wallet_1 = wallet_node_1.wallet_state_manager.main_wallet

        ph0 = await wallet_0.get_new_puzzlehash()
        ph1 = await wallet_1.get_new_puzzlehash()

        sk0 = await wallet_node_0.wallet_state_manager.get_private_key(ph0)
        sk1 = await wallet_node_1.wallet_state_manager.get_private_key(ph1)
        assert sk0 == sk1

        if trusted:
            wallet_node_0.config["trusted_peers"] = {
                full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
            }
            wallet_node_1.config["trusted_peers"] = {
                full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
            }

        else:
            wallet_node_0.config["trusted_peers"] = {}
            wallet_node_1.config["trusted_peers"] = {}
        await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
        await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

        await full_node_api.farm_blocks_to_wallet(1, wallet_0)
        await full_node_api.farm_blocks_to_wallet(1, wallet_1)

        # Wallet1 sets up DIDWallet1 without any backup set
        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_0: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager, wallet_0, uint64(101), action_scope
            )

        with pytest.raises(RuntimeError):
            assert await did_wallet_0.get_coin() == set()
        assert await did_wallet_0.get_info_for_recovery() is None

        await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
        await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node_0, wallet_node_1])

        await time_out_assert(15, did_wallet_0.get_confirmed_balance, 101)
        await time_out_assert(15, did_wallet_0.get_unconfirmed_balance, 101)
        await time_out_assert(15, did_wallet_0.get_pending_change_balance, 0)

        await full_node_api.farm_blocks_to_wallet(1, wallet_0)

        #######################
        all_node_0_wallets = await wallet_node_0.wallet_state_manager.user_store.get_all_wallet_info_entries()
        all_node_1_wallets = await wallet_node_1.wallet_state_manager.user_store.get_all_wallet_info_entries()
        assert (
            json.loads(all_node_0_wallets[1].data)["current_inner"]
            == json.loads(all_node_1_wallets[1].data)["current_inner"]
        )

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize(
        "wallet_environments",
        [
            {
                "num_environments": 3,
                "blocks_needed": [1, 1, 1],
            }
        ],
        indirect=True,
    )
    @pytest.mark.anyio
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    async def test_creation_from_backup_file(self, wallet_environments: WalletTestFramework) -> None:
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        env_2 = wallet_environments.environments[2]

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        env_1.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        env_2.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        # Wallet1 sets up DIDWallet1 without any backup set
        async with env_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_0: DIDWallet = await DIDWallet.create_new_did_wallet(
                env_0.wallet_state_manager, env_0.xck_wallet, uint64(101), action_scope
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "unconfirmed_wallet_balance": -101,
                            "<=#spendable_balance": -101,
                            "<=#max_send_amount": -101,
                            ">=#pending_change": 1,
                            "pending_coin_removal_count": 1,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": -101,
                            ">=#spendable_balance": 1,
                            ">=#max_send_amount": 1,
                            "<=#pending_change": -1,
                            "pending_coin_removal_count": -1,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Wallet1 sets up DIDWallet_1 with DIDWallet_0 as backup
        backup_ids = [bytes32.from_hexstr(did_wallet_0.get_my_DID())]

        async with env_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                env_1.wallet_state_manager, env_1.xck_wallet, uint64(201), action_scope, backup_ids
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "unconfirmed_wallet_balance": -201,
                            "<=#spendable_balance": -201,
                            "<=#max_send_amount": -201,
                            ">=#pending_change": 1,
                            "pending_coin_removal_count": 1,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 201,
                            "pending_change": 201,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": -201,
                            ">=#spendable_balance": 1,
                            ">=#max_send_amount": 1,
                            "<=#pending_change": -1,
                            "pending_coin_removal_count": -1,
                        },
                        "did": {
                            "confirmed_wallet_balance": 201,
                            "spendable_balance": 201,
                            "max_send_amount": 201,
                            "unspent_coin_count": 1,
                            "pending_change": -201,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        backup_data = did_wallet_1.create_backup()

        # Wallet2 recovers DIDWallet2 to a new set of keys
        await env_2.rpc_client.create_new_did_wallet(
            uint64(1),
            DEFAULT_TX_CONFIG,
            type="recovery",
            backup_data=backup_data,
        )
        did_wallet_2 = env_2.wallet_state_manager.get_wallet(id=uint32(2), required_type=DIDWallet)
        recovery_info = await env_2.rpc_client.did_get_recovery_info(
            DIDGetRecoveryInfo(uint32(env_2.wallet_aliases["did"]))
        )
        assert recovery_info.wallet_id == env_2.wallet_aliases["did"]
        assert recovery_info.backup_dids == backup_ids
        current_coin_info_response = await env_0.rpc_client.did_get_current_coin_info(
            DIDGetCurrentCoinInfo(uint32(env_0.wallet_aliases["did"]))
        )
        # TODO: this check is kind of weak, we should research when this endpoint might actually be useful
        assert current_coin_info_response.wallet_id == env_0.wallet_aliases["did"]
        async with env_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            message_spend_bundle, attest_data = await did_wallet_0.create_attestment(
                recovery_info.coin_name, recovery_info.newpuzhash, recovery_info.pubkey, action_scope
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "spendable_balance": -101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                            "max_send_amount": -101,
                        }
                    },
                    post_block_balance_updates={
                        "did": {
                            "spendable_balance": 101,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                            "max_send_amount": 101,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "init": True,
                        }
                    },
                    post_block_balance_updates={},
                ),
            ]
        )

        (
            test_info_list,
            test_message_spend_bundle,
        ) = await did_wallet_2.load_attest_files_for_recovery_spend([attest_data])
        assert message_spend_bundle == test_message_spend_bundle

        async with env_2.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            assert did_wallet_2.did_info.temp_coin is not None
            await did_wallet_2.recovery_spend(
                did_wallet_2.did_info.temp_coin,
                recovery_info.newpuzhash,
                test_info_list,
                recovery_info.pubkey,
                test_message_spend_bundle,
                action_scope,
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        "did": {
                            "confirmed_wallet_balance": -201,
                            "unconfirmed_wallet_balance": -201,
                            "spendable_balance": -201,
                            "max_send_amount": -201,
                            "unspent_coin_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "unconfirmed_wallet_balance": 201,
                            "pending_coin_removal_count": 2,
                        }
                    },
                    post_block_balance_updates={
                        "did": {
                            "confirmed_wallet_balance": 201,
                            "spendable_balance": 201,
                            "max_send_amount": 201,
                            "unspent_coin_count": 1,
                            "pending_coin_removal_count": -2,
                        }
                    },
                ),
            ]
        )

        for wallet in [did_wallet_0, did_wallet_1, did_wallet_2]:
            assert wallet.wallet_state_manager.wallets[wallet.id()] == wallet

        some_ph = bytes32(32 * b"\2")
        async with env_2.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet_2.create_exit_spend(some_ph, action_scope)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "unconfirmed_wallet_balance": -201,
                            "spendable_balance": -201,
                            "max_send_amount": -201,
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        "did": {
                            "confirmed_wallet_balance": -201,
                            "unspent_coin_count": -1,
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
            ]
        )

        async def get_coins_with_ph() -> bool:
            coins = await wallet_environments.full_node.full_node.coin_store.get_coin_records_by_puzzle_hash(
                True, some_ph
            )
            return len(coins) == 1

        await time_out_assert(15, get_coins_with_ph, True)

        for wallet in [did_wallet_0, did_wallet_1]:
            assert wallet.wallet_state_manager.wallets[wallet.id()] == wallet

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 2, "blocks_needed": [1, 1]}], indirect=True)
    @pytest.mark.anyio
    async def test_did_recovery_with_multiple_backup_dids(self, wallet_environments: WalletTestFramework):
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_node_0 = env_0.node
        wallet_node_1 = env_1.node
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        env_1.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager, wallet_0, uint64(101), action_scope
            )
        assert did_wallet.get_name() == "Profile 1"
        recovery_list = [bytes32.from_hexstr(did_wallet.get_my_DID())]

        async with wallet_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_2: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_1.wallet_state_manager, wallet_1, uint64(101), action_scope, recovery_list
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )
        assert did_wallet_2.did_info.backup_ids == recovery_list

        recovery_list.append(bytes32.from_hexstr(did_wallet_2.get_my_DID()))

        async with wallet_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_3: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_1.wallet_state_manager, wallet_1, uint64(201), action_scope, recovery_list
            )

        env_1.wallet_aliases["did_2"] = 3

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "set_remainder": True,
                        },
                        "did_2": {
                            "init": True,
                            "unconfirmed_wallet_balance": 201,
                            "pending_change": 201,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "set_remainder": True,
                        },
                        "did_2": {
                            "confirmed_wallet_balance": 201,
                            "spendable_balance": 201,
                            "max_send_amount": 201,
                            "unspent_coin_count": 1,
                            "pending_change": -201,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )
        coin = await did_wallet_3.get_coin()

        backup_data = did_wallet_3.create_backup()

        async with wallet_node_0.wallet_state_manager.lock:
            did_wallet_4 = await DIDWallet.create_new_did_wallet_from_recovery(
                wallet_node_0.wallet_state_manager,
                wallet_0,
                backup_data,
            )
        assert did_wallet_4.get_name() == "Profile 2"
        env_0.wallet_aliases["did_2"] = 3

        pubkey = (
            await did_wallet_4.wallet_state_manager.get_unused_derivation_record(did_wallet_2.wallet_info.id)
        ).pubkey
        new_ph = did_wallet_4.did_info.temp_puzhash
        async with did_wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            message_spend_bundle, attest1 = await did_wallet.create_attestment(
                coin.name(), new_ph, pubkey, action_scope
            )

        async with did_wallet_2.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope_2:
            message_spend_bundle2, attest2 = await did_wallet_2.create_attestment(
                coin.name(), new_ph, pubkey, action_scope_2
            )

        message_spend_bundle = message_spend_bundle.aggregate([message_spend_bundle, message_spend_bundle2])

        (
            test_info_list,
            test_message_spend_bundle,
        ) = await did_wallet_4.load_attest_files_for_recovery_spend([attest1, attest2])
        assert message_spend_bundle == test_message_spend_bundle

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "spendable_balance": -101,
                            "pending_change": 101,
                            "max_send_amount": -101,
                            "pending_coin_removal_count": 1,
                            "set_remainder": True,
                        },
                        "did_2": {
                            "init": True,
                            "unconfirmed_wallet_balance": 0,
                            "pending_change": 0,
                            "pending_coin_removal_count": 0,
                        },
                    },
                    post_block_balance_updates={
                        "did": {
                            "spendable_balance": 101,
                            "pending_change": -101,
                            "max_send_amount": 101,
                            "pending_coin_removal_count": -1,
                        },
                        "did_2": {
                            "confirmed_wallet_balance": 0,
                            "spendable_balance": 0,
                            "max_send_amount": 0,
                            "unspent_coin_count": 0,
                            "pending_change": 0,
                            "pending_coin_removal_count": 0,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "spendable_balance": -101,
                            "pending_change": 101,
                            "max_send_amount": -101,
                            "pending_coin_removal_count": 1,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did": {
                            "spendable_balance": 101,
                            "pending_change": -101,
                            "max_send_amount": 101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )

        async with did_wallet_4.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet_4.recovery_spend(coin, new_ph, test_info_list, pubkey, message_spend_bundle, action_scope)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_2": {
                            "unconfirmed_wallet_balance": 201,
                            "pending_change": 0,
                            "pending_coin_removal_count": 3,
                        },
                    },
                    post_block_balance_updates={
                        "did_2": {
                            "confirmed_wallet_balance": 201,
                            "spendable_balance": 201,
                            "max_send_amount": 201,
                            "unspent_coin_count": 1,
                            "pending_change": 0,
                            "pending_coin_removal_count": -3,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_2": {
                            "unconfirmed_wallet_balance": 0,  # TODO: fix pre-block balances for recovery
                            "spendable_balance": 0,
                            "pending_change": 0,
                            "max_send_amount": 0,
                            "pending_coin_removal_count": 0,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did_2": {
                            "confirmed_wallet_balance": -201,
                            "spendable_balance": -201,
                            "max_send_amount": -201,
                            "unspent_coin_count": -1,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )

        for wallet in [did_wallet, did_wallet_2, did_wallet_3, did_wallet_4]:
            assert wallet.wallet_state_manager.wallets[wallet.id()] == wallet

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 1, "blocks_needed": [1]}], indirect=True)
    @pytest.mark.anyio
    async def test_did_recovery_with_empty_set(self, wallet_environments: WalletTestFramework):
        env_0 = wallet_environments.environments[0]
        wallet_node_0 = env_0.node
        wallet_0 = env_0.xck_wallet

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        ph = await wallet_0.get_new_puzzlehash()

        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager, wallet_0, uint64(101), action_scope
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )
        coin = await did_wallet.get_coin()
        info: list[tuple[bytes, bytes, int]] = []
        pubkey = (await did_wallet.wallet_state_manager.get_unused_derivation_record(did_wallet.wallet_info.id)).pubkey
        with pytest.raises(Exception):  # We expect a KLVM 80 error for this test
            async with did_wallet.wallet_state_manager.new_action_scope(
                wallet_environments.tx_config, push=False
            ) as action_scope:
                await did_wallet.recovery_spend(
                    coin,
                    ph,
                    info,
                    pubkey,
                    WalletSpendBundle([], AugSchemeMPL.aggregate([])),
                    action_scope,
                )

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 1, "blocks_needed": [1]}], indirect=True)
    @pytest.mark.anyio
    async def test_did_find_lost_did(self, wallet_environments: WalletTestFramework):
        env_0 = wallet_environments.environments[0]
        wallet_node_0 = env_0.node
        wallet_0 = env_0.xck_wallet
        api_0 = env_0.rpc_api

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager, wallet_0, uint64(101), action_scope
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )

        # Delete the coin and wallet
        coin = await did_wallet.get_coin()
        await wallet_node_0.wallet_state_manager.coin_store.delete_coin_record(coin.name())
        await wallet_node_0.wallet_state_manager.delete_wallet(did_wallet.wallet_info.id)
        wallet_node_0.wallet_state_manager.wallets.pop(did_wallet.wallet_info.id)
        assert len(wallet_node_0.wallet_state_manager.wallets) == 1
        # Find lost DID
        assert did_wallet.did_info.origin_coin is not None  # mypy
        resp = await api_0.did_find_lost_did({"coin_id": did_wallet.did_info.origin_coin.name().hex()})
        assert resp["success"]
        did_wallets = list(
            filter(
                lambda w: (w.type == WalletType.DECENTRALIZED_ID),
                await wallet_node_0.wallet_state_manager.get_all_wallet_info_entries(),
            )
        )
        did_wallet = wallet_node_0.wallet_state_manager.wallets[did_wallets[0].id]
        env_0.wallet_aliases["did_found"] = did_wallets[0].id
        await env_0.change_balances(
            {
                "did_found": {
                    "init": True,
                    "confirmed_wallet_balance": 101,
                    "unconfirmed_wallet_balance": 101,
                    "spendable_balance": 101,
                    "max_send_amount": 101,
                    "unspent_coin_count": 1,
                }
            }
        )
        await env_0.check_balances()

        # Spend DID
        recovery_list = [bytes32.fromhex(did_wallet.get_my_DID())]
        await did_wallet.update_recovery_list(recovery_list, uint64(1))
        assert did_wallet.did_info.backup_ids == recovery_list
        async with did_wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet.create_update_spend(action_scope)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_found": {
                            "spendable_balance": -101,
                            "max_send_amount": -101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "did_found": {
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )

        # Delete the coin and change inner puzzle
        coin = await did_wallet.get_coin()
        await wallet_node_0.wallet_state_manager.coin_store.delete_coin_record(coin.name())
        new_inner_puzzle = await did_wallet.get_did_innerpuz(new=True)
        did_wallet.did_info = dataclasses.replace(did_wallet.did_info, current_inner=new_inner_puzzle)
        # Recovery the coin
        assert did_wallet.did_info.origin_coin is not None  # mypy
        resp = await api_0.did_find_lost_did({"coin_id": did_wallet.did_info.origin_coin.name().hex()})
        assert resp["success"]
        found_coin = await did_wallet.get_coin()
        assert found_coin == coin
        assert did_wallet.did_info.current_inner != new_inner_puzzle

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 2, "blocks_needed": [1, 1]}], indirect=True)
    @pytest.mark.anyio
    async def test_did_attest_after_recovery(self, wallet_environments: WalletTestFramework):
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_node_0 = env_0.node
        wallet_node_1 = env_1.node
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        env_1.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager, wallet_0, uint64(101), action_scope
            )
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
                WalletStateTransition(),
            ]
        )

        await time_out_assert(15, did_wallet.get_confirmed_balance, 101)
        await time_out_assert(15, did_wallet.get_unconfirmed_balance, 101)
        recovery_list = [bytes32.from_hexstr(did_wallet.get_my_DID())]

        async with wallet_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_2: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_1.wallet_state_manager, wallet_1, uint64(101), action_scope, recovery_list
            )
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )

        assert did_wallet_2.did_info.backup_ids == recovery_list

        # Update coin with new ID info
        recovery_list = [bytes32.from_hexstr(did_wallet_2.get_my_DID())]
        await did_wallet.update_recovery_list(recovery_list, uint64(1))
        assert did_wallet.did_info.backup_ids == recovery_list
        async with did_wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet.create_update_spend(action_scope)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        }
                    },
                    post_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        }
                    },
                ),
                WalletStateTransition(),
            ]
        )

        # DID Wallet 2 recovers into DID Wallet 3 with new innerpuz
        backup_data = did_wallet_2.create_backup()

        async with wallet_node_0.wallet_state_manager.lock:
            did_wallet_3 = await DIDWallet.create_new_did_wallet_from_recovery(
                wallet_node_0.wallet_state_manager,
                wallet_0,
                backup_data,
            )
        env_0.wallet_aliases["did_2"] = 3
        new_ph = await did_wallet_3.get_did_inner_hash(new=True)
        coin = await did_wallet_2.get_coin()
        pubkey = (
            await did_wallet_3.wallet_state_manager.get_unused_derivation_record(did_wallet_3.wallet_info.id)
        ).pubkey

        async with did_wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            message_spend_bundle, attest_data = await did_wallet.create_attestment(
                coin.name(), new_ph, pubkey, action_scope
            )
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        },
                        "did_2": {
                            "init": True,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        },
                        "did_2": {
                            "set_remainder": True,
                        },
                    },
                ),
                WalletStateTransition(),
            ]
        )

        (
            info,
            message_spend_bundle,
        ) = await did_wallet_3.load_attest_files_for_recovery_spend([attest_data])
        async with did_wallet_3.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await did_wallet_3.recovery_spend(coin, new_ph, info, pubkey, message_spend_bundle, action_scope)
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_2": {
                            "unconfirmed_wallet_balance": 101,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did_2": {
                            "confirmed_wallet_balance": 101,
                            "set_remainder": True,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "unconfirmed_wallet_balance": 0,  # TODO: fix pre-block balances for recovery
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did": {
                            "confirmed_wallet_balance": -101,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )

        # DID Wallet 1 recovery spends into DID Wallet 4
        backup_data = did_wallet.create_backup()

        async with wallet_node_1.wallet_state_manager.lock:
            did_wallet_4 = await DIDWallet.create_new_did_wallet_from_recovery(
                wallet_node_1.wallet_state_manager,
                wallet_1,
                backup_data,
            )
        env_1.wallet_aliases["did_2"] = 3
        coin = await did_wallet.get_coin()
        new_ph = await did_wallet_4.get_did_inner_hash(new=True)
        pubkey = (
            await did_wallet_4.wallet_state_manager.get_unused_derivation_record(did_wallet_4.wallet_info.id)
        ).pubkey
        async with did_wallet_3.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            message_spend_bundle, attest1 = await did_wallet_3.create_attestment(
                coin.name(), new_ph, pubkey, action_scope
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_2": {
                            "unconfirmed_wallet_balance": 0,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did_2": {
                            "confirmed_wallet_balance": 0,
                            "set_remainder": True,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_2": {
                            "init": True,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did_2": {
                            "confirmed_wallet_balance": 0,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )

        (
            test_info_list,
            test_message_spend_bundle,
        ) = await did_wallet_4.load_attest_files_for_recovery_spend([attest1])
        async with did_wallet_4.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await did_wallet_4.recovery_spend(
                coin, new_ph, test_info_list, pubkey, test_message_spend_bundle, action_scope
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did": {
                            "confirmed_wallet_balance": -101,
                            "set_remainder": True,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did_2": {
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did_2": {
                            "confirmed_wallet_balance": 101,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )

        for wallet in [did_wallet, did_wallet_3, did_wallet_4]:
            assert wallet.wallet_state_manager.wallets[wallet.id()] == wallet

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 2, "blocks_needed": [1, 1]}], indirect=True)
    @pytest.mark.parametrize(
        "with_recovery",
        [True, False],
    )
    @pytest.mark.anyio
    async def test_did_transfer(self, wallet_environments: WalletTestFramework, with_recovery: bool):
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_node_0 = env_0.node
        wallet_node_1 = env_1.node
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        env_1.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        ph = await wallet_0.get_new_puzzlehash()
        fee = uint64(1000)

        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager,
                wallet_0,
                uint64(101),
                action_scope,
                [ph],
                uint64(1),
                {"Twitter": "Test", "GitHub": "测试"},
                fee=fee,
            )
        assert did_wallet_1.get_name() == "Profile 1"

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
                WalletStateTransition(),
            ]
        )

        # Transfer DID
        new_puzhash = await wallet_1.get_new_puzzlehash()
        async with did_wallet_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet_1.transfer_did(new_puzhash, fee, with_recovery, action_scope)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "unconfirmed_wallet_balance": -101,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        "did": {
                            "init": True,
                            "confirmed_wallet_balance": 101,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )

        # Get the new DID wallet
        did_wallets = list(
            filter(
                lambda w: (w.type == WalletType.DECENTRALIZED_ID),
                await wallet_node_1.wallet_state_manager.get_all_wallet_info_entries(),
            )
        )
        did_wallet_2 = wallet_node_1.wallet_state_manager.wallets[did_wallets[0].id]
        assert isinstance(did_wallet_2, DIDWallet)  # mypy
        assert len(wallet_node_0.wallet_state_manager.wallets) == 1
        assert did_wallet_1.did_info.origin_coin == did_wallet_2.did_info.origin_coin
        if with_recovery:
            assert did_wallet_1.did_info.backup_ids[0] == did_wallet_2.did_info.backup_ids[0]
            assert did_wallet_1.did_info.num_of_backup_ids_needed == did_wallet_2.did_info.num_of_backup_ids_needed
        metadata = json.loads(did_wallet_2.did_info.metadata)
        assert metadata["Twitter"] == "Test"
        assert metadata["GitHub"] == "测试"

        # Test match_hinted_coin
        assert await did_wallet_2.match_hinted_coin(
            await did_wallet_2.get_coin(),
            new_puzhash,
        )

    @pytest.mark.parametrize(
        "trusted",
        [True, False],
    )
    @pytest.mark.anyio
    async def test_did_auto_transfer_limit(
        self,
        self_hostname: str,
        two_wallet_nodes: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
        trusted: bool,
    ) -> None:
        fee = uint64(1000)
        full_nodes, wallets, _ = two_wallet_nodes
        full_node_api = full_nodes[0]
        server_1 = full_node_api.server
        wallet_node, server_2 = wallets[0]
        wallet_node_2, server_3 = wallets[1]
        wallet = wallet_node.wallet_state_manager.main_wallet
        wallet2 = wallet_node_2.wallet_state_manager.main_wallet
        api_1 = WalletRpcApi(wallet_node_2)
        ph = await wallet.get_new_puzzlehash()

        if trusted:
            wallet_node.config["trusted_peers"] = {
                full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
            }
            wallet_node_2.config["trusted_peers"] = {
                full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
            }
        else:
            wallet_node.config["trusted_peers"] = {}
            wallet_node_2.config["trusted_peers"] = {}

        await server_2.start_client(PeerInfo(self_hostname, server_1.get_port()), None)
        await server_3.start_client(PeerInfo(self_hostname, server_1.get_port()), None)
        await full_node_api.farm_blocks_to_wallet(1, wallet)

        # Check that we cap out at 10 DID Wallets automatically created upon transfer received
        for i in range(0, 14):
            async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
                did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                    wallet_node.wallet_state_manager,
                    wallet,
                    uint64(101),
                    action_scope,
                    [bytes32(bytes(ph))],
                    uint64(1),
                    {"Twitter": "Test", "GitHub": "测试"},
                    fee=fee,
                )
            assert did_wallet_1.get_name() == "Profile 1"
            await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
            await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node, wallet_node_2])
            await time_out_assert(15, did_wallet_1.get_confirmed_balance, 101)
            await time_out_assert(15, did_wallet_1.get_unconfirmed_balance, 101)
            # Transfer DID
            assert did_wallet_1.did_info.origin_coin is not None
            origin_coin = did_wallet_1.did_info.origin_coin
            new_puzhash = await wallet2.get_new_puzzlehash()
            async with did_wallet_1.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
                await did_wallet_1.transfer_did(new_puzhash, fee, False, action_scope)
            await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
            await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node, wallet_node_2])
            # Check if the DID wallet is created in the wallet2

            await time_out_assert(
                30, get_wallet_num, min(2 + i, 11), wallet_node_2.wallet_state_manager
            )  # check we haven't made more than 10 DID wallets
            await time_out_assert(30, get_wallet_num, 1, wallet_node.wallet_state_manager)
        # Get the new DID wallets
        did_wallets = list(
            filter(
                lambda w: (w.type == WalletType.DECENTRALIZED_ID),
                await wallet_node_2.wallet_state_manager.get_all_wallet_info_entries(),
            )
        )

        assert len(did_wallets) == 10
        # Test we can use the DID
        did_wallet_10 = wallet_node_2.wallet_state_manager.get_wallet(
            id=uint32(did_wallets[9].id), required_type=DIDWallet
        )
        # Delete the coin and change inner puzzle
        coin = await did_wallet_10.get_coin()
        # origin_coin = did_wallet_10.did_info.origin_coin
        backup_data = did_wallet_10.create_backup()
        await wallet_node_2.wallet_state_manager.coin_store.delete_coin_record(coin.name())
        await time_out_assert(15, did_wallet_10.get_confirmed_balance, 0)
        await wallet_node_2.wallet_state_manager.user_store.delete_wallet(did_wallet_10.wallet_info.id)
        wallet_node_2.wallet_state_manager.wallets.pop(did_wallet_10.wallet_info.id)
        # Recover the coin
        async with wallet_node_2.wallet_state_manager.lock:
            did_wallet_10 = await DIDWallet.create_new_did_wallet_from_recovery(
                wallet_node_2.wallet_state_manager,
                wallet2,
                backup_data,
            )
        assert did_wallet_10.did_info.origin_coin is not None
        resp = await api_1.did_find_lost_did({"coin_id": did_wallet_10.did_info.origin_coin.name().hex()})
        assert resp["success"]
        await time_out_assert(15, did_wallet_10.get_confirmed_balance, 101)
        await time_out_assert(15, did_wallet_10.get_unconfirmed_balance, 101)

        # Check we can recover an auto-discarded DID
        did_wallet_9 = wallet_node_2.wallet_state_manager.get_wallet(
            id=uint32(did_wallets[8].id), required_type=DIDWallet
        )
        # Delete the coin and wallet to make space for a auto-discarded DID
        coin = await did_wallet_9.get_coin()
        await wallet_node_2.wallet_state_manager.coin_store.delete_coin_record(coin.name())
        await time_out_assert(15, did_wallet_9.get_confirmed_balance, 0)
        await wallet_node_2.wallet_state_manager.user_store.delete_wallet(did_wallet_9.wallet_info.id)
        wallet_node_2.wallet_state_manager.wallets.pop(did_wallet_9.wallet_info.id)

        did_wallets = list(
            filter(
                lambda w: (w.type == WalletType.DECENTRALIZED_ID),
                await wallet_node_2.wallet_state_manager.get_all_wallet_info_entries(),
            )
        )
        assert len(did_wallets) == 9

        # Try and find lost coin
        resp = await api_1.did_find_lost_did({"coin_id": origin_coin.name().hex()})
        did_wallets = list(
            filter(
                lambda w: (w.type == WalletType.DECENTRALIZED_ID),
                await wallet_node_2.wallet_state_manager.get_all_wallet_info_entries(),
            )
        )
        assert len(did_wallets) == 10

        # Check we can still manually add new DIDs while at cap
        await full_node_api.farm_blocks_to_wallet(1, wallet2)
        ph = await wallet2.get_new_puzzlehash()
        async with wallet2.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_11: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_2.wallet_state_manager,
                wallet2,
                uint64(101),
                action_scope,
                [bytes32(bytes(ph))],
                uint64(1),
                {"Twitter": "Test", "GitHub": "测试"},
                fee=fee,
            )
        await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
        await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node, wallet_node_2])
        await time_out_assert(15, did_wallet_11.get_confirmed_balance, 101)
        await time_out_assert(15, did_wallet_11.get_unconfirmed_balance, 101)

        did_wallets = list(
            filter(
                lambda w: (w.type == WalletType.DECENTRALIZED_ID),
                await wallet_node_2.wallet_state_manager.get_all_wallet_info_entries(),
            )
        )
        assert len(did_wallets) == 11

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 1, "blocks_needed": [1]}], indirect=True)
    @pytest.mark.anyio
    async def test_update_recovery_list(self, wallet_environments: WalletTestFramework):
        env = wallet_environments.environments[0]
        wallet_node = env.node
        wallet = env.xck_wallet

        env.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        ph = await wallet.get_new_puzzlehash()

        async with wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node.wallet_state_manager, wallet, uint64(101), action_scope, []
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )
        await did_wallet_1.update_recovery_list([ph], uint64(1))
        async with did_wallet_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet_1.create_update_spend(action_scope)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "did": {
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )
        assert did_wallet_1.did_info.backup_ids[0] == bytes(ph)
        assert did_wallet_1.did_info.num_of_backup_ids_needed == 1

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 2, "blocks_needed": [1, 1]}], indirect=True)
    @pytest.mark.anyio
    async def test_get_info(self, wallet_environments: WalletTestFramework):
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_node_0 = env_0.node
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet
        api_0 = env_0.rpc_api

        env_0.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        env_1.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        fee = uint64(1000)
        did_amount = uint64(101)
        ph_1 = await wallet_1.get_new_puzzlehash()

        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager,
                wallet_0,
                did_amount,
                action_scope,
                [],
                metadata={"twitter": "twitter"},
                fee=fee,
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": did_amount,
                            "pending_change": did_amount,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": did_amount,
                            "spendable_balance": did_amount,
                            "max_send_amount": did_amount,
                            "unspent_coin_count": 1,
                            "pending_change": -did_amount,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
                WalletStateTransition(),
            ]
        )
        assert did_wallet_1.did_info.origin_coin is not None  # mypy
        response = await api_0.did_get_info({"coin_id": did_wallet_1.did_info.origin_coin.name().hex()})
        assert response["did_id"] == encode_puzzle_hash(did_wallet_1.did_info.origin_coin.name(), AddressType.DID.value)
        assert response["launcher_id"] == did_wallet_1.did_info.origin_coin.name().hex()
        assert did_wallet_1.did_info.current_inner is not None  # mypy
        assert response["full_puzzle"].to_program() == create_singleton_puzzle(
            did_wallet_1.did_info.current_inner, did_wallet_1.did_info.origin_coin.name()
        )
        assert response["metadata"]["twitter"] == "twitter"
        assert response["latest_coin"] == (await did_wallet_1.get_coin()).name().hex()
        assert response["num_verification"] == 0
        assert response["recovery_list_hash"] == Program(Program.to([])).get_tree_hash().hex()
        assert decode_puzzle_hash(response["p2_address"]).hex() == response["hints"][0]

        # Test non-singleton coin
        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            coin = (await wallet_0.select_coins(uint64(1), action_scope)).pop()
        assert coin.amount % 2 == 1
        coin_id = coin.name()
        response = await api_0.did_get_info({"coin_id": coin_id.hex()})
        assert not response["success"]

        # Test multiple odd coins
        odd_amount = uint64(1)
        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            async with action_scope.use() as interface:
                interface.side_effects.selected_coins.append(coin)
            coin_1 = (await wallet_0.select_coins(odd_amount, action_scope)).pop()
        assert coin_1.amount % 2 == 0
        async with wallet_0.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config.override(excluded_coin_ids=[coin_id]), push=True
        ) as action_scope:
            await wallet_0.generate_signed_transaction(odd_amount, ph_1, action_scope, fee)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "unconfirmed_wallet_balance": -odd_amount - fee,
                            "set_remainder": True,
                        },
                        "did": {
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": -odd_amount - fee,
                            "set_remainder": True,
                        },
                        "did": {
                            "set_remainder": True,
                        },
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "unconfirmed_wallet_balance": 0,
                            "set_remainder": True,
                        }
                    },
                    post_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": odd_amount,
                            "set_remainder": True,
                        }
                    },
                ),
            ]
        )

        with pytest.raises(ValueError):
            await api_0.did_get_info({"coin_id": coin_1.name().hex()})

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 1, "blocks_needed": [1]}], indirect=True)
    @pytest.mark.anyio
    async def test_message_spend(self, wallet_environments: WalletTestFramework):
        env = wallet_environments.environments[0]
        wallet_node = env.node
        wallet = env.xck_wallet
        api_0 = env.rpc_api

        env.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        fee = uint64(1000)

        async with wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node.wallet_state_manager, wallet, uint64(101), action_scope, [], fee=fee
            )
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )
        response = await api_0.did_message_spend(
            {"wallet_id": did_wallet_1.wallet_id, "coin_announcements": ["0abc"], "puzzle_announcements": ["0def"]}
        )
        spend = response["spend_bundle"].coin_spends[0]
        conditions = conditions_dict_for_solution(
            spend.puzzle_reveal, spend.solution, wallet.wallet_state_manager.constants.MAX_BLOCK_COST_KLVM
        )

        assert len(conditions[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT]) == 1
        assert conditions[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT][0].vars[0].hex() == "0abc"
        assert len(conditions[ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT]) == 1
        assert conditions[ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT][0].vars[0].hex() == "0def"

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 1, "blocks_needed": [1]}], indirect=True)
    @pytest.mark.anyio
    async def test_update_metadata(self, wallet_environments: WalletTestFramework):
        env = wallet_environments.environments[0]
        wallet_node = env.node
        wallet = env.xck_wallet

        env.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }

        fee = uint64(1000)
        did_amount = uint64(101)

        async with wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node.wallet_state_manager, wallet, did_amount, action_scope, [], fee=fee
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )

        assert did_wallet_1.did_info.current_inner is not None  # mypy
        puzhash = did_wallet_1.did_info.current_inner.get_tree_hash()
        parent_num = get_parent_num(did_wallet_1)

        bad_metadata = {"Twitter": {"url": "http://www.twitter.com"}}
        with pytest.raises(ValueError) as e:
            await did_wallet_1.update_metadata(bad_metadata)  # type: ignore
        assert e.match("Metadata key value pairs must be strings.")

        metadata = {}
        metadata["Twitter"] = "http://www.twitter.com"
        await did_wallet_1.update_metadata(metadata)
        async with did_wallet_1.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet_1.create_update_spend(action_scope, fee)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "unconfirmed_wallet_balance": -fee,
                            "set_remainder": True,
                        },
                        "did": {
                            "unconfirmed_wallet_balance": 0,
                            "set_remainder": True,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": -fee,
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 0,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )

        assert get_parent_num(did_wallet_1) == parent_num + 2
        assert did_wallet_1.did_info.current_inner is not None  # mypy
        assert puzhash != did_wallet_1.did_info.current_inner.get_tree_hash()
        assert did_wallet_1.did_info.metadata.find("Twitter") > 0

    @pytest.mark.limit_consensus_modes(allowed=[ConsensusMode.PLAIN], reason="irrelevant")
    @pytest.mark.parametrize("wallet_environments", [{"num_environments": 1, "blocks_needed": [1]}], indirect=True)
    @pytest.mark.anyio
    async def test_did_sign_message(self, wallet_environments: WalletTestFramework):
        env = wallet_environments.environments[0]
        wallet_node = env.node
        wallet = env.xck_wallet
        api_0 = env.rpc_api

        env.wallet_aliases = {
            "xck": 1,
            "did": 2,
        }
        fee = uint64(1000)
        ph = await wallet.get_new_puzzlehash()

        async with wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node.wallet_state_manager,
                wallet,
                uint64(101),
                action_scope,
                [ph],
                uint64(1),
                {"Twitter": "Test", "GitHub": "测试"},
                fee=fee,
            )
        assert did_wallet_1.get_name() == "Profile 1"

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "init": True,
                            "unconfirmed_wallet_balance": 101,
                            "pending_change": 101,
                            "pending_coin_removal_count": 1,
                        },
                    },
                    post_block_balance_updates={
                        "xck": {
                            "set_remainder": True,
                        },
                        "did": {
                            "confirmed_wallet_balance": 101,
                            "spendable_balance": 101,
                            "max_send_amount": 101,
                            "unspent_coin_count": 1,
                            "pending_change": -101,
                            "pending_coin_removal_count": -1,
                        },
                    },
                ),
            ]
        )
        # Test general string
        assert did_wallet_1.did_info.origin_coin is not None  # mypy
        message = "Hello World"
        assert did_wallet_1.did_info.origin_coin is not None
        response = await api_0.sign_message_by_id(
            {
                "id": encode_puzzle_hash(did_wallet_1.did_info.origin_coin.name(), AddressType.DID.value),
                "message": message,
            }
        )
        puzzle: Program = Program.to((CHIP_0002_SIGN_MESSAGE_PREFIX, message))
        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            puzzle.get_tree_hash(),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )
        # Test hex string
        message = "0123456789ABCDEF"
        response = await api_0.sign_message_by_id(
            {
                "id": encode_puzzle_hash(did_wallet_1.did_info.origin_coin.name(), AddressType.DID.value),
                "message": message,
                "is_hex": True,
            }
        )
        puzzle = Program.to((CHIP_0002_SIGN_MESSAGE_PREFIX, bytes.fromhex(message)))

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            puzzle.get_tree_hash(),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )

        # Test BLS sign string
        message = "Hello World"
        assert did_wallet_1.did_info.origin_coin is not None
        response = await api_0.sign_message_by_id(
            {
                "id": encode_puzzle_hash(did_wallet_1.did_info.origin_coin.name(), AddressType.DID.value),
                "message": message,
                "is_hex": "False",
                "safe_mode": "False",
            }
        )

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            bytes(message, "utf-8"),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )
        # Test BLS sign hex
        message = "0123456789ABCDEF"
        assert did_wallet_1.did_info.origin_coin is not None
        response = await api_0.sign_message_by_id(
            {
                "id": encode_puzzle_hash(did_wallet_1.did_info.origin_coin.name(), AddressType.DID.value),
                "message": message,
                "is_hex": True,
                "safe_mode": False,
            }
        )

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            bytes.fromhex(message),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )

    @pytest.mark.parametrize(
        "trusted",
        [True, False],
    )
    @pytest.mark.anyio
    async def test_create_did_with_recovery_list(
        self, self_hostname: str, two_nodes_two_wallets_with_same_keys: OldSimulatorsAndWallets, trusted: bool
    ) -> None:
        """
        A DID is created on-chain in client0, causing a DID Wallet to be created in client1, which shares the same key.
        This can happen if someone uses the same key on multiple computers, or is syncing a wallet from scratch.

        For this test, we assign a recovery list hash at DID creation time, but the recovery list is not yet available
        to the wallet_node that the DID Wallet is being created in (client1).

        """
        full_nodes, wallets, _ = two_nodes_two_wallets_with_same_keys
        full_node_api = full_nodes[0]
        full_node_server = full_node_api.server
        wallet_node_0, server_0 = wallets[0]
        wallet_node_1, server_1 = wallets[1]

        wallet_0 = wallet_node_0.wallet_state_manager.main_wallet
        wallet_1 = wallet_node_1.wallet_state_manager.main_wallet

        ph0 = await wallet_0.get_new_puzzlehash()
        ph1 = await wallet_1.get_new_puzzlehash()

        sk0 = await wallet_node_0.wallet_state_manager.get_private_key(ph0)
        sk1 = await wallet_node_1.wallet_state_manager.get_private_key(ph1)
        assert sk0 == sk1

        if trusted:
            wallet_node_0.config["trusted_peers"] = {
                full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
            }
            wallet_node_1.config["trusted_peers"] = {
                full_node_api.full_node.server.node_id.hex(): full_node_api.full_node.server.node_id.hex()
            }

        else:
            wallet_node_0.config["trusted_peers"] = {}
            wallet_node_1.config["trusted_peers"] = {}
        await server_0.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)
        await server_1.start_client(PeerInfo(self_hostname, full_node_server.get_port()), None)

        await full_node_api.farm_blocks_to_wallet(1, wallet_0)
        await full_node_api.farm_blocks_to_wallet(1, wallet_1)

        # Node 0 sets up a DID Wallet with a backup set, but num_of_backup_ids_needed=0
        # (a malformed solution, but legal for the klvm puzzle)
        recovery_list = [bytes32(bytes.fromhex("00" * 32))]
        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_0: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_0.wallet_state_manager,
                wallet_0,
                uint64(101),
                action_scope,
                backups_ids=recovery_list,
                num_of_backup_ids_needed=uint64(0),
            )

        await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
        await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node_0])

        await time_out_assert(15, did_wallet_0.get_confirmed_balance, 101)
        await time_out_assert(15, did_wallet_0.get_unconfirmed_balance, 101)
        await time_out_assert(15, did_wallet_0.get_pending_change_balance, 0)

        await full_node_api.farm_blocks_to_wallet(1, wallet_0)

        #######################
        all_node_0_wallets = await wallet_node_0.wallet_state_manager.user_store.get_all_wallet_info_entries()
        all_node_1_wallets = await wallet_node_1.wallet_state_manager.user_store.get_all_wallet_info_entries()
        assert len(all_node_0_wallets) == len(all_node_1_wallets)

        # Note that the inner program we expect is different than the on-chain inner.
        # This means that we have more work to do in the checks for the two different spend cases of
        # the DID wallet Singleton
        # assert (
        #    json.loads(all_node_0_wallets[1].data)["current_inner"]
        #    == json.loads(all_node_1_wallets[1].data)["current_inner"]
        # )

    #  TODO: See Issue CHIK-1544
    #  This test should be ported to WalletTestFramework once we can replace keys in the wallet node
    @pytest.mark.parametrize(
        "trusted",
        [True, False],
    )
    @pytest.mark.anyio
    async def test_did_resync(
        self,
        self_hostname: str,
        two_wallet_nodes: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
        trusted: bool,
    ) -> None:
        full_nodes, wallets, _ = two_wallet_nodes
        full_node_api = full_nodes[0]
        full_node_server = full_node_api.full_node.server
        wallet_node_1, wallet_server_1 = wallets[0]
        wallet_node_2, wallet_server_2 = wallets[1]
        wallet = wallet_node_1.wallet_state_manager.main_wallet
        wallet2 = wallet_node_2.wallet_state_manager.main_wallet
        fee = uint64(0)
        wallet_api_1 = WalletRpcApi(wallet_node_1)
        wallet_api_2 = WalletRpcApi(wallet_node_2)
        ph = await wallet.get_new_puzzlehash()
        if trusted:
            wallet_node_1.config["trusted_peers"] = {full_node_server.node_id.hex(): full_node_server.node_id.hex()}
            wallet_node_2.config["trusted_peers"] = {full_node_server.node_id.hex(): full_node_server.node_id.hex()}
        else:
            wallet_node_1.config["trusted_peers"] = {}
            wallet_node_2.config["trusted_peers"] = {}
        assert full_node_server._port is not None
        await wallet_server_1.start_client(PeerInfo(self_hostname, uint16(full_node_server._port)), None)
        await wallet_server_2.start_client(PeerInfo(self_hostname, uint16(full_node_server._port)), None)
        await full_node_api.farm_blocks_to_wallet(1, wallet)

        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            did_wallet_1: DIDWallet = await DIDWallet.create_new_did_wallet(
                wallet_node_1.wallet_state_manager,
                wallet,
                uint64(101),
                action_scope,
                [bytes32(ph)],
                uint64(1),
                {"Twitter": "Test", "GitHub": "测试"},
                fee=fee,
            )
        assert did_wallet_1.get_name() == "Profile 1"
        await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
        await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node_1, wallet_node_2])
        await time_out_assert(15, did_wallet_1.get_confirmed_balance, 101)
        await time_out_assert(15, did_wallet_1.get_unconfirmed_balance, 101)
        # Transfer DID
        new_puzhash = await wallet2.get_new_puzzlehash()
        async with did_wallet_1.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await did_wallet_1.transfer_did(new_puzhash, fee, True, action_scope=action_scope)
        await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
        await full_node_api.wait_for_wallets_synced(wallet_nodes=[wallet_node_1, wallet_node_2])
        # Check if the DID wallet is created in the wallet2
        await time_out_assert(30, get_wallet_num, 2, wallet_node_2.wallet_state_manager)
        await time_out_assert(30, get_wallet_num, 1, wallet_node_1.wallet_state_manager)
        did_wallet_2 = wallet_node_2.wallet_state_manager.get_wallet(uint32(2), DIDWallet)
        did_info = did_wallet_2.did_info
        # set flag to reset wallet sync data on start
        await wallet_api_1.set_wallet_resync_on_startup({"enable": True})
        fingerprint_1 = wallet_node_1.logged_in_fingerprint
        await wallet_api_2.set_wallet_resync_on_startup({"enable": True})
        fingerprint_2 = wallet_node_2.logged_in_fingerprint
        # 2 reward coins
        assert len(await wallet_node_1.wallet_state_manager.coin_store.get_all_unspent_coins()) == 2
        # Delete tx records
        await wallet_node_1.wallet_state_manager.tx_store.rollback_to_block(0)
        wallet_node_1._close()
        await wallet_node_1._await_closed()
        wallet_node_2._close()
        await wallet_node_2._await_closed()
        wallet_node_1.config["database_path"] = "wallet/db/blockchain_wallet_v2_test_1_CHALLENGE_KEY.sqlite"
        wallet_node_2.config["database_path"] = "wallet/db/blockchain_wallet_v2_test_2_CHALLENGE_KEY.sqlite"
        # Start resync
        await wallet_node_1._start_with_fingerprint(fingerprint_1)
        await wallet_node_2._start_with_fingerprint(fingerprint_2)
        assert full_node_server._port is not None
        await wallet_server_1.start_client(PeerInfo(self_hostname, uint16(full_node_server._port)), None)
        await wallet_server_2.start_client(PeerInfo(self_hostname, uint16(full_node_server._port)), None)
        await full_node_api.farm_new_transaction_block(FarmNewBlockProtocol(bytes32(b"\00" * 32)))
        await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_1, timeout=20)
        await full_node_api.wait_for_wallet_synced(wallet_node=wallet_node_2, timeout=20)
        await time_out_assert(30, get_wallet_num, 1, wallet_node_1.wallet_state_manager)
        await time_out_assert(30, get_wallet_num, 2, wallet_node_2.wallet_state_manager)
        did_wallet_2 = wallet_node_2.wallet_state_manager.get_wallet(uint32(2), DIDWallet)
        assert did_info == did_wallet_2.did_info


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_did_coin_records(wallet_environments: WalletTestFramework, monkeypatch: pytest.MonkeyPatch) -> None:
    # Setup
    wallet_node = wallet_environments.environments[0].node
    wallet = wallet_environments.environments[0].xck_wallet

    # Generate DID wallet
    async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
        did_wallet: DIDWallet = await DIDWallet.create_new_did_wallet(
            wallet_node.wallet_state_manager, wallet, uint64(1), action_scope
        )

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    1: {"set_remainder": True},
                    2: {"init": True, "set_remainder": True},
                },
                post_block_balance_updates={
                    1: {"set_remainder": True},
                    2: {"set_remainder": True},
                },
            ),
            WalletStateTransition(),
        ]
    )

    for _ in range(0, 2):
        async with did_wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=True
        ) as action_scope:
            await did_wallet.transfer_did(await wallet.get_puzzle_hash(new=False), uint64(0), True, action_scope)
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {"set_remainder": True},
                        2: {"set_remainder": True},
                    },
                    post_block_balance_updates={
                        1: {"set_remainder": True},
                        2: {"set_remainder": True},
                    },
                ),
                WalletStateTransition(),
            ]
        )

    assert len(await wallet.wallet_state_manager.get_spendable_coins_for_wallet(did_wallet.id())) == 1



from __future__ import annotations

import os
import stat
import sys
from logging import Logger
from pathlib import Path
from typing import Optional

from chik.util.config import load_config, traverse_dict
from chik.util.permissions import octal_mode_string, verify_file_permissions

DEFAULT_PERMISSIONS_CERT_FILE: int = 0o644
DEFAULT_PERMISSIONS_KEY_FILE: int = 0o600

# Masks containing permission bits we don't allow
RESTRICT_MASK_CERT_FILE: int = stat.S_IWGRP | stat.S_IXGRP | stat.S_IWOTH | stat.S_IXOTH  # 0o033
RESTRICT_MASK_KEY_FILE: int = (
    stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH
)  # 0o077

CERT_CONFIG_KEY_PATHS = [
    "chik_ssl_ca:crt",
    "daemon_ssl:private_crt",
    "farmer:ssl:private_crt",
    "farmer:ssl:public_crt",
    "full_node:ssl:private_crt",
    "full_node:ssl:public_crt",
    "data_layer:ssl:private_crt",
    "data_layer:ssl:public_crt",
    "harvester:chik_ssl_ca:crt",
    "harvester:private_ssl_ca:crt",
    "harvester:ssl:private_crt",
    "introducer:ssl:public_crt",
    "private_ssl_ca:crt",
    "timelord:ssl:private_crt",
    "timelord:ssl:public_crt",
    "ui:daemon_ssl:private_crt",
    "wallet:ssl:private_crt",
    "wallet:ssl:public_crt",
]
KEY_CONFIG_KEY_PATHS = [
    "chik_ssl_ca:key",
    "daemon_ssl:private_key",
    "farmer:ssl:private_key",
    "farmer:ssl:public_key",
    "full_node:ssl:private_key",
    "full_node:ssl:public_key",
    "harvester:chik_ssl_ca:key",
    "harvester:private_ssl_ca:key",
    "harvester:ssl:private_key",
    "introducer:ssl:public_key",
    "private_ssl_ca:key",
    "timelord:ssl:private_key",
    "timelord:ssl:public_key",
    "ui:daemon_ssl:private_key",
    "wallet:ssl:private_key",
    "wallet:ssl:public_key",
]


# Set to keep track of which files we've already warned about
warned_ssl_files: set[Path] = set()


def get_all_ssl_file_paths(root_path: Path) -> tuple[list[Path], list[Path]]:
    """Lookup config values and append to a list of files whose permissions we need to check"""
    from chik.ssl.create_ssl import get_mozilla_ca_crt

    all_certs: list[Path] = []
    all_keys: list[Path] = []

    try:
        config = load_config(root_path, "config.yaml", exit_on_error=False, fill_missing_services=True)
        for paths, parsed_list in [(CERT_CONFIG_KEY_PATHS, all_certs), (KEY_CONFIG_KEY_PATHS, all_keys)]:
            for path in paths:
                try:
                    file = root_path / Path(traverse_dict(config, path))
                    parsed_list.append(file)
                except Exception as e:
                    print(
                        f"Failed to lookup config value for {path}: {e}"
                    )  # lgtm [py/clear-text-logging-sensitive-data]

        # Check the Mozilla Root CAs as well
        all_certs.append(Path(get_mozilla_ca_crt()))
    except (FileNotFoundError, ValueError):
        pass

    return all_certs, all_keys


def get_ssl_perm_warning(path: Path, actual_mode: int, expected_mode: int) -> str:
    return (
        f"Permissions {octal_mode_string(actual_mode)} for "
        f"'{path}' are too open. "  # lgtm [py/clear-text-logging-sensitive-data]
        f"Expected {octal_mode_string(expected_mode)}"
    )


def verify_ssl_certs_and_keys(
    cert_paths: list[Path], key_paths: list[Path], log: Optional[Logger] = None
) -> list[tuple[Path, int, int]]:
    """Check that file permissions are properly set for the provided SSL cert and key files"""
    if sys.platform == "win32" or sys.platform == "cygwin":
        # TODO: ACLs for SSL certs/keys on Windows
        return []

    invalid_files_and_modes: list[tuple[Path, int, int]] = []

    def verify_paths(paths: list[Path], restrict_mask: int, expected_permissions: int) -> None:
        nonlocal invalid_files_and_modes
        for path in paths:
            try:
                # Check that the file permissions are not too permissive
                is_valid, actual_permissions = verify_file_permissions(path, restrict_mask)
                if not is_valid:
                    if log is not None:
                        log.error(get_ssl_perm_warning(path, actual_permissions, expected_permissions))
                    warned_ssl_files.add(path)
                    invalid_files_and_modes.append((path, actual_permissions, expected_permissions))
            except FileNotFoundError:
                # permissions can't be dangerously wrong on nonexistent files
                pass
            except Exception as e:
                print(f"Unable to check permissions for {path}: {e}")  # lgtm [py/clear-text-logging-sensitive-data]

    verify_paths(cert_paths, RESTRICT_MASK_CERT_FILE, DEFAULT_PERMISSIONS_CERT_FILE)
    verify_paths(key_paths, RESTRICT_MASK_KEY_FILE, DEFAULT_PERMISSIONS_KEY_FILE)

    return invalid_files_and_modes


def check_ssl(root_path: Path) -> None:
    """
    Sanity checks on the SSL configuration. Checks that file permissions are properly
    set on the keys and certs, warning and exiting if permissions are incorrect.
    """
    if sys.platform == "win32" or sys.platform == "cygwin":
        # TODO: ACLs for SSL certs/keys on Windows
        return None

    certs_to_check, keys_to_check = get_all_ssl_file_paths(root_path)
    invalid_files = verify_ssl_certs_and_keys(certs_to_check, keys_to_check)
    if len(invalid_files):
        lines = [
            "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@",
            "@             WARNING: UNPROTECTED SSL FILE!              @",
            "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@",
            *(
                get_ssl_perm_warning(path, actual_permissions, expected_permissions)
                for path, actual_permissions, expected_permissions in invalid_files
            ),
            "One or more SSL files were found with permission issues.",
            "Run the following to fix issues: chik init --fix-ssl-permissions",
        ]
        print("\n".join(lines), file=sys.stderr)


def check_and_fix_permissions_for_ssl_file(file: Path, mask: int, updated_mode: int) -> tuple[bool, bool]:
    """Check file permissions and attempt to fix them if found to be too open"""
    if sys.platform == "win32" or sys.platform == "cygwin":
        # TODO: ACLs for SSL certs/keys on Windows
        return True, False

    valid: bool = True
    updated: bool = False

    # Check that the file permissions are not too permissive
    try:
        (good_perms, _mode) = verify_file_permissions(file, mask)
        if not good_perms:
            valid = False
            print(
                f"Attempting to set permissions {octal_mode_string(updated_mode)} on "
                f"{file}"  # lgtm [py/clear-text-logging-sensitive-data]
            )
            os.chmod(str(file), updated_mode)
            updated = True
    except Exception as e:
        print(f"Failed to change permissions on {file}: {e}")  # lgtm [py/clear-text-logging-sensitive-data]
        valid = False

    return valid, updated


def fix_ssl(root_path: Path) -> None:
    """Attempts to fix SSL cert/key file permissions that are too open"""

    if sys.platform == "win32" or sys.platform == "cygwin":
        # TODO: ACLs for SSL certs/keys on Windows
        return None

    updated: bool = False
    encountered_error: bool = False

    certs_to_check, keys_to_check = get_all_ssl_file_paths(root_path)
    files_to_fix = verify_ssl_certs_and_keys(certs_to_check, keys_to_check)

    for file, mask, updated_mode in files_to_fix:
        # Check that permissions are correct, and if not, attempt to fix
        (valid, fixed) = check_and_fix_permissions_for_ssl_file(file, mask, updated_mode)
        if fixed:
            updated = True
        if not valid and not fixed:
            encountered_error = True

    if encountered_error:
        print("One or more errors were encountered while updating SSL file permissions...")
    elif updated:
        print("Finished updating SSL file permissions")
    else:
        print("SSL file permissions are correct")



from __future__ import annotations

import logging
import os
from logging.handlers import SysLogHandler
from pathlib import Path
from typing import Any, Optional, cast

import colorlog
from concurrent_log_handler import ConcurrentRotatingFileHandler

from chik import __version__
from chik.util.chik_version import chik_short_version
from chik.util.path import path_from_root

default_log_level = "WARNING"


def get_beta_logging_config() -> dict[str, Any]:
    return {
        "log_filename": f"{chik_short_version()}/chik-blockchain/beta.log",
        "log_level": "DEBUG",
        "log_stdout": False,
        "log_maxfilesrotation": 100,
        "log_maxbytesrotation": 100 * 1024 * 1024,
        "log_use_gzip": True,
    }


def get_file_log_handler(
    formatter: logging.Formatter, root_path: Path, logging_config: dict[str, object]
) -> ConcurrentRotatingFileHandler:
    log_path = path_from_root(root_path, str(logging_config.get("log_filename", "log/debug.log")))
    log_path.parent.mkdir(parents=True, exist_ok=True)
    maxrotation = cast(int, logging_config.get("log_maxfilesrotation", 7))
    maxbytesrotation = cast(int, logging_config.get("log_maxbytesrotation", 50 * 1024 * 1024))
    use_gzip = cast(bool, logging_config.get("log_use_gzip", False))
    handler = ConcurrentRotatingFileHandler(
        os.fspath(log_path), "a", maxBytes=maxbytesrotation, backupCount=maxrotation, use_gzip=use_gzip
    )
    handler.setFormatter(formatter)
    return handler


def initialize_logging(
    service_name: str,
    logging_config: dict[str, Any],
    root_path: Path,
    beta_root_path: Optional[Path] = None,
) -> None:
    log_backcompat = logging_config.get("log_backcompat", False)
    log_level = logging_config.get("log_level", default_log_level)
    file_name_length = 33 - len(service_name)
    log_date_format = "%Y-%m-%dT%H:%M:%S"
    file_log_formatter = logging.Formatter(
        fmt=(
            f"%(asctime)s.%(msecs)03d {service_name} %(name)-{file_name_length}s: %(levelname)-8s %(message)s"
            if log_backcompat
            else f"%(asctime)s.%(msecs)03d {__version__} {service_name} %(name)-{file_name_length}s: "
            f"%(levelname)-8s %(message)s"
        ),
        datefmt=log_date_format,
    )
    handlers: list[logging.Handler] = []
    if logging_config["log_stdout"]:
        stdout_handler = colorlog.StreamHandler()
        stdout_handler.setFormatter(
            colorlog.ColoredFormatter(
                (
                    f"%(asctime)s.%(msecs)03d {service_name} %(name)-{file_name_length}s: "
                    f"%(log_color)s%(levelname)-8s%(reset)s %(message)s"
                    if log_backcompat
                    else f"%(asctime)s.%(msecs)03d {__version__} {service_name} %(name)-{file_name_length}s: "
                    f"%(log_color)s%(levelname)-8s%(reset)s %(message)s"
                ),
                datefmt=log_date_format,
                reset=True,
            )
        )
        handlers.append(stdout_handler)
    else:
        handlers.append(get_file_log_handler(file_log_formatter, root_path, logging_config))

    if logging_config.get("log_syslog", False):
        log_syslog_host = logging_config.get("log_syslog_host", "localhost")
        log_syslog_port = logging_config.get("log_syslog_port", 514)
        log_syslog_handler = SysLogHandler(address=(log_syslog_host, log_syslog_port))
        log_syslog_handler.setFormatter(logging.Formatter(fmt=f"{service_name} %(message)s", datefmt=log_date_format))
        handlers.append(log_syslog_handler)

    if beta_root_path is not None:
        handlers.append(get_file_log_handler(file_log_formatter, beta_root_path, get_beta_logging_config()))

    root_logger = logging.getLogger()
    for handler in handlers:
        root_logger.addHandler(handler)

    set_log_level(log_level=log_level, service_name=service_name)


def set_log_level(log_level: str, service_name: str) -> list[str]:
    root_logger = logging.getLogger()
    log_level_exceptions = {}

    for handler in root_logger.handlers:
        try:
            handler.setLevel(log_level)
        except Exception as e:
            handler.setLevel(default_log_level)
            log_level_exceptions[handler] = e

    error_strings = [
        f"Handler {handler}: Invalid log level '{log_level}' for {service_name}. "
        f"Defaulting to: {default_log_level}. Error: {exception}"
        for handler, exception in log_level_exceptions.items()
    ]
    for error_string in error_strings:
        root_logger.error(error_string)

    # Adjust the root logger to the smallest used log level since its default level is WARNING which would overwrite
    # the potentially smaller log levels of specific handlers.
    root_logger.setLevel(min(handler.level for handler in root_logger.handlers))

    if root_logger.level <= logging.DEBUG:
        logging.getLogger("aiosqlite").setLevel(logging.INFO)  # Too much logging on debug level

    return error_strings


def initialize_service_logging(service_name: str, config: dict[str, Any], root_path: Path) -> None:
    if service_name == "daemon":
        # TODO: Maybe introduce a separate `daemon` section in the config instead of having `daemon_port`, `logging`
        #  and the daemon related stuff as top level entries.
        logging_config = config["logging"]
    else:
        logging_config = config[service_name]["logging"]
    beta_config = config.get("beta", {})
    beta_config_path = beta_config.get("path") if beta_config.get("enabled", False) else None
    initialize_logging(
        service_name=service_name,
        logging_config=logging_config,
        root_path=root_path,
        beta_root_path=beta_config_path,
    )



from __future__ import annotations

import dataclasses
from dataclasses import dataclass, field
from enum import Enum, IntEnum
from hashlib import sha256
from typing import TYPE_CHECKING, Any, Optional, Union

import aiosqlite
from typing_extensions import final

from chik.data_layer.data_layer_errors import ProofIntegrityError
from chik.server.ws_connection import WSChikConnection
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.byte_types import hexstr_to_bytes
from chik.util.db_wrapper import DBWrapper2
from chik.util.ints import uint8, uint64
from chik.util.streamable import Streamable, streamable
from chik.wallet.db_wallet.db_wallet_puzzles import create_host_fullpuz

if TYPE_CHECKING:
    from chik.data_layer.data_store import DataStore
    from chik.wallet.wallet_node import WalletNode


def internal_hash(left_hash: bytes32, right_hash: bytes32) -> bytes32:
    # see test for the definition this is optimized from
    return bytes32(sha256(b"\2" + left_hash + right_hash).digest())


def calculate_internal_hash(hash: bytes32, other_hash_side: Side, other_hash: bytes32) -> bytes32:
    if other_hash_side == Side.LEFT:
        return internal_hash(left_hash=other_hash, right_hash=hash)
    elif other_hash_side == Side.RIGHT:
        return internal_hash(left_hash=hash, right_hash=other_hash)

    raise Exception(f"Invalid side: {other_hash_side!r}")


def leaf_hash(key: bytes, value: bytes) -> bytes32:
    # see test for the definition this is optimized from
    return bytes32(sha256(b"\2" + sha256(b"\1" + key).digest() + sha256(b"\1" + value).digest()).digest())


def key_hash(key: bytes) -> bytes32:
    # see test for the definition this is optimized from
    return bytes32(sha256(b"\1" + key).digest())


@dataclasses.dataclass(frozen=True)
class PaginationData:
    total_pages: int
    total_bytes: int
    hashes: list[bytes32]


def get_hashes_for_page(page: int, lengths: dict[bytes32, int], max_page_size: int) -> PaginationData:
    current_page = 0
    current_page_size = 0
    total_bytes = 0
    hashes: list[bytes32] = []
    for hash, length in sorted(lengths.items(), key=lambda x: (-x[1], x[0])):
        if length > max_page_size:
            raise RuntimeError(
                f"Cannot paginate data, item size is larger than max page size: {length} {max_page_size}"
            )
        total_bytes += length
        if current_page_size + length <= max_page_size:
            current_page_size += length
        else:
            current_page += 1
            current_page_size = length
        if current_page == page:
            hashes.append(hash)

    return PaginationData(current_page + 1, total_bytes, hashes)


async def _debug_dump(db: DBWrapper2, description: str = "") -> None:
    async with db.reader() as reader:
        cursor = await reader.execute("SELECT name FROM sqlite_master WHERE type='table';")
        print("-" * 50, description, flush=True)
        for [name] in await cursor.fetchall():
            cursor = await reader.execute(f"SELECT * FROM {name}")
            print(f"\n -- {name} ------", flush=True)
            async for row in cursor:
                print(f"        {dict(row)}")


async def _dot_dump(
    data_store: DataStore,
    store_id: bytes32,
    root_hash: bytes32,
) -> str:
    terminal_nodes = await data_store.get_keys_values(store_id=store_id, root_hash=root_hash)
    internal_nodes = await data_store.get_internal_nodes(store_id=store_id, root_hash=root_hash)

    n = 8

    dot_nodes: list[str] = []
    dot_connections: list[str] = []
    dot_pair_boxes: list[str] = []

    for terminal_node in terminal_nodes:
        hash = terminal_node.hash.hex()
        key = terminal_node.key.hex()
        value = terminal_node.value.hex()
        dot_nodes.append(f"""node_{hash} [shape=box, label="{hash[:n]}\\nkey: {key}\\nvalue: {value}"];""")

    for internal_node in internal_nodes:
        hash = internal_node.hash.hex()
        left = internal_node.left_hash.hex()
        right = internal_node.right_hash.hex()
        dot_nodes.append(f"""node_{hash} [label="{hash[:n]}"]""")
        dot_connections.append(f"""node_{hash} -> node_{left} [label="L"];""")
        dot_connections.append(f"""node_{hash} -> node_{right} [label="R"];""")
        dot_pair_boxes.append(
            f"node [shape = box]; {{rank = same; node_{left}->node_{right}[style=invis]; rankdir = LR}}"
        )

    lines = [
        "digraph {",
        *dot_nodes,
        *dot_connections,
        *dot_pair_boxes,
        "}",
    ]

    return "\n".join(lines)


def row_to_node(row: aiosqlite.Row) -> Node:
    cls = node_type_to_class[row["node_type"]]
    return cls.from_row(row=row)


class Status(IntEnum):
    PENDING = 1
    COMMITTED = 2
    PENDING_BATCH = 3


class NodeType(IntEnum):
    INTERNAL = 1
    TERMINAL = 2


@final
class Side(IntEnum):
    LEFT = 0
    RIGHT = 1

    def other(self) -> Side:
        if self == Side.LEFT:
            return Side.RIGHT

        return Side.LEFT

    @classmethod
    def unmarshal(cls, o: str) -> Side:
        return getattr(cls, o.upper())  # type: ignore[no-any-return]

    def marshal(self) -> str:
        return self.name.lower()


class OperationType(IntEnum):
    INSERT = 0
    DELETE = 1


class CommitState(IntEnum):
    OPEN = 0
    FINALIZED = 1
    ROLLED_BACK = 2


Node = Union["TerminalNode", "InternalNode"]


@final
@dataclass(frozen=True)
class TerminalNode:
    hash: bytes32
    # generation: int
    key: bytes
    value: bytes

    # left for now for interface back-compat even though it is constant
    atom: None = field(init=False, default=None)

    @classmethod
    def from_key_value(cls, key: bytes, value: bytes) -> TerminalNode:
        return cls(
            hash=leaf_hash(key=key, value=value),
            key=key,
            value=value,
        )

    @classmethod
    def from_row(cls, row: aiosqlite.Row) -> TerminalNode:
        return cls(
            hash=bytes32(row["hash"]),
            # generation=row["generation"],
            key=row["key"],
            value=row["value"],
        )


@final
@dataclass(frozen=True)
class ProofOfInclusionLayer:
    other_hash_side: Side
    other_hash: bytes32
    combined_hash: bytes32

    @classmethod
    def from_internal_node(
        cls,
        internal_node: InternalNode,
        traversal_child_hash: bytes32,
    ) -> ProofOfInclusionLayer:
        return ProofOfInclusionLayer(
            other_hash_side=internal_node.other_child_side(hash=traversal_child_hash),
            other_hash=internal_node.other_child_hash(hash=traversal_child_hash),
            combined_hash=internal_node.hash,
        )

    @classmethod
    def from_hashes(cls, primary_hash: bytes32, other_hash_side: Side, other_hash: bytes32) -> ProofOfInclusionLayer:
        combined_hash = calculate_internal_hash(
            hash=primary_hash,
            other_hash_side=other_hash_side,
            other_hash=other_hash,
        )

        return cls(other_hash_side=other_hash_side, other_hash=other_hash, combined_hash=combined_hash)


other_side_to_bit = {Side.LEFT: 1, Side.RIGHT: 0}


@dataclass(frozen=True)
class ProofOfInclusion:
    node_hash: bytes32
    # children before parents
    layers: list[ProofOfInclusionLayer]

    @property
    def root_hash(self) -> bytes32:
        if len(self.layers) == 0:
            return self.node_hash

        return self.layers[-1].combined_hash

    def sibling_sides_integer(self) -> int:
        return sum(other_side_to_bit[layer.other_hash_side] << index for index, layer in enumerate(self.layers))

    def sibling_hashes(self) -> list[bytes32]:
        return [layer.other_hash for layer in self.layers]

    def as_program(self) -> Program:
        return Program.to([self.sibling_sides_integer(), self.sibling_hashes()])

    def valid(self) -> bool:
        existing_hash = self.node_hash

        for layer in self.layers:
            calculated_hash = calculate_internal_hash(
                hash=existing_hash, other_hash_side=layer.other_hash_side, other_hash=layer.other_hash
            )

            if calculated_hash != layer.combined_hash:
                return False

            existing_hash = calculated_hash

        if existing_hash != self.root_hash:
            return False

        return True


@final
@dataclass(frozen=True)
class InternalNode:
    hash: bytes32
    # generation: int
    left_hash: bytes32
    right_hash: bytes32

    left: Optional[Node] = None
    right: Optional[Node] = None

    @classmethod
    def from_child_nodes(cls, left: Node, right: Node) -> InternalNode:
        return cls(
            hash=internal_hash(left_hash=left.hash, right_hash=right.hash),
            left_hash=left.hash,
            right_hash=right.hash,
            left=left,
            right=right,
        )

    @classmethod
    def from_row(cls, row: aiosqlite.Row) -> InternalNode:
        return cls(
            hash=bytes32(row["hash"]),
            # generation=row["generation"],
            left_hash=bytes32(row["left"]),
            right_hash=bytes32(row["right"]),
        )

    def other_child_hash(self, hash: bytes32) -> bytes32:
        if self.left_hash == hash:
            return self.right_hash
        elif self.right_hash == hash:
            return self.left_hash

        # TODO: real exception considerations
        raise Exception("provided hash not present")

    def other_child_side(self, hash: bytes32) -> Side:
        if self.left_hash == hash:
            return Side.RIGHT
        elif self.right_hash == hash:
            return Side.LEFT

        # TODO: real exception considerations
        raise Exception("provided hash not present")


class Unspecified(Enum):
    # not beautiful, improve when a better way is known
    # https://github.com/python/typing/issues/236#issuecomment-229515556

    instance = None

    def __repr__(self) -> str:
        return "Unspecified"


unspecified = Unspecified.instance


@dataclass(frozen=True)
class Root:
    store_id: bytes32
    node_hash: Optional[bytes32]
    generation: int
    status: Status

    @classmethod
    def from_row(cls, row: aiosqlite.Row) -> Root:
        raw_node_hash = row["node_hash"]
        if raw_node_hash is None:
            node_hash = None
        else:
            node_hash = bytes32(raw_node_hash)

        return cls(
            store_id=bytes32(row["tree_id"]),
            node_hash=node_hash,
            generation=row["generation"],
            status=Status(row["status"]),
        )

    def to_row(self) -> dict[str, Any]:
        return {
            "tree_id": self.store_id,
            "node_hash": self.node_hash,
            "generation": self.generation,
            "status": self.status.value,
        }

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> Root:
        return cls(
            store_id=bytes32.from_hexstr(marshalled["tree_id"]),
            node_hash=None if marshalled["node_hash"] is None else bytes32.from_hexstr(marshalled["node_hash"]),
            generation=marshalled["generation"],
            status=Status(marshalled["status"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "tree_id": self.store_id.hex(),
            "node_hash": None if self.node_hash is None else self.node_hash.hex(),
            "generation": self.generation,
            "status": self.status.value,
        }


node_type_to_class: dict[NodeType, Union[type[InternalNode], type[TerminalNode]]] = {
    NodeType.INTERNAL: InternalNode,
    NodeType.TERMINAL: TerminalNode,
}


@dataclass(frozen=True)
class ServerInfo:
    url: str
    num_consecutive_failures: int
    ignore_till: int


@dataclass(frozen=True)
class Subscription:
    store_id: bytes32
    servers_info: list[ServerInfo]


@dataclass(frozen=True)
class DiffData:
    type: OperationType
    key: bytes
    value: bytes


@streamable
@dataclass(frozen=True)
class SerializedNode(Streamable):
    is_terminal: bool
    value1: bytes
    value2: bytes


@final
@dataclasses.dataclass(frozen=True)
class KeyValue:
    key: bytes
    value: bytes

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> KeyValue:
        return cls(
            key=hexstr_to_bytes(marshalled["key"]),
            value=hexstr_to_bytes(marshalled["value"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "key": self.key.hex(),
            "value": self.value.hex(),
        }


@dataclasses.dataclass(frozen=True)
class OfferStore:
    store_id: bytes32
    inclusions: tuple[KeyValue, ...]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> OfferStore:
        return cls(
            store_id=bytes32.from_hexstr(marshalled["store_id"]),
            inclusions=tuple(KeyValue.unmarshal(key_value) for key_value in marshalled["inclusions"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "store_id": self.store_id.hex(),
            "inclusions": [key_value.marshal() for key_value in self.inclusions],
        }


@dataclasses.dataclass(frozen=True)
class Layer:
    # This class is similar to chik.data_layer.data_layer_util.ProofOfInclusionLayer
    # but is being retained for now to keep the API schema definition localized here.

    other_hash_side: Side
    other_hash: bytes32
    combined_hash: bytes32

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> Layer:
        return cls(
            other_hash_side=Side.unmarshal(marshalled["other_hash_side"]),
            other_hash=bytes32.from_hexstr(marshalled["other_hash"]),
            combined_hash=bytes32.from_hexstr(marshalled["combined_hash"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "other_hash_side": self.other_hash_side.marshal(),
            "other_hash": self.other_hash.hex(),
            "combined_hash": self.combined_hash.hex(),
        }


@dataclasses.dataclass(frozen=True)
class MakeOfferRequest:
    maker: tuple[OfferStore, ...]
    taker: tuple[OfferStore, ...]
    fee: Optional[uint64]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> MakeOfferRequest:
        return cls(
            maker=tuple(OfferStore.unmarshal(offer_store) for offer_store in marshalled["maker"]),
            taker=tuple(OfferStore.unmarshal(offer_store) for offer_store in marshalled["taker"]),
            fee=None if marshalled["fee"] is None else uint64(marshalled["fee"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "maker": [offer_store.marshal() for offer_store in self.maker],
            "taker": [offer_store.marshal() for offer_store in self.taker],
            "fee": None if self.fee is None else int(self.fee),
        }


@dataclasses.dataclass(frozen=True)
class Proof:
    key: bytes
    value: bytes
    node_hash: bytes32
    layers: tuple[Layer, ...]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> Proof:
        return cls(
            key=hexstr_to_bytes(marshalled["key"]),
            value=hexstr_to_bytes(marshalled["value"]),
            node_hash=bytes32.from_hexstr(marshalled["node_hash"]),
            layers=tuple(Layer.unmarshal(layer) for layer in marshalled["layers"]),
        )

    def root(self) -> bytes32:
        if len(self.layers) == 0:
            return self.node_hash

        return self.layers[-1].combined_hash

    def marshal(self) -> dict[str, Any]:
        return {
            "key": self.key.hex(),
            "value": self.value.hex(),
            "node_hash": self.node_hash.hex(),
            "layers": [layer.marshal() for layer in self.layers],
        }


@dataclasses.dataclass(frozen=True)
class StoreProofs:
    store_id: bytes32
    proofs: tuple[Proof, ...]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> StoreProofs:
        return cls(
            store_id=bytes32.from_hexstr(marshalled["store_id"]),
            proofs=tuple(Proof.unmarshal(proof) for proof in marshalled["proofs"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "store_id": self.store_id.hex(),
            "proofs": [proof.marshal() for proof in self.proofs],
        }


@dataclasses.dataclass(frozen=True)
class Offer:
    trade_id: bytes
    offer: bytes
    taker: tuple[OfferStore, ...]
    maker: tuple[StoreProofs, ...]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> Offer:
        return cls(
            trade_id=bytes32.from_hexstr(marshalled["trade_id"]),
            offer=hexstr_to_bytes(marshalled["offer"]),
            taker=tuple(OfferStore.unmarshal(offer_store) for offer_store in marshalled["taker"]),
            maker=tuple(StoreProofs.unmarshal(store_proof) for store_proof in marshalled["maker"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "trade_id": self.trade_id.hex(),
            "offer": self.offer.hex(),
            "taker": [offer_store.marshal() for offer_store in self.taker],
            "maker": [store_proofs.marshal() for store_proofs in self.maker],
        }


@dataclasses.dataclass(frozen=True)
class MakeOfferResponse:
    success: bool
    offer: Offer

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> MakeOfferResponse:
        return cls(
            success=marshalled["success"],
            offer=Offer.unmarshal(marshalled["offer"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "success": self.success,
            "offer": self.offer.marshal(),
        }


@dataclasses.dataclass(frozen=True)
class TakeOfferRequest:
    offer: Offer
    fee: Optional[uint64]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> TakeOfferRequest:
        return cls(
            offer=Offer.unmarshal(marshalled["offer"]),
            fee=None if marshalled["fee"] is None else uint64(marshalled["fee"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "offer": self.offer.marshal(),
            "fee": None if self.fee is None else int(self.fee),
        }


@dataclasses.dataclass(frozen=True)
class TakeOfferResponse:
    success: bool
    trade_id: bytes32

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> TakeOfferResponse:
        return cls(
            success=marshalled["success"],
            trade_id=bytes32.from_hexstr(marshalled["trade_id"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "success": self.success,
            "trade_id": self.trade_id.hex(),
        }


@final
@dataclasses.dataclass(frozen=True)
class VerifyOfferResponse:
    success: bool
    valid: bool
    error: Optional[str] = None
    fee: Optional[uint64] = None

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> VerifyOfferResponse:
        return cls(
            success=marshalled["success"],
            valid=marshalled["valid"],
            error=marshalled["error"],
            fee=None if marshalled["fee"] is None else uint64(marshalled["fee"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "success": self.success,
            "valid": self.valid,
            "error": self.error,
            "fee": None if self.fee is None else int(self.fee),
        }


@dataclasses.dataclass(frozen=True)
class CancelOfferRequest:
    trade_id: bytes32
    # cancel on chain (secure) vs. just locally
    secure: bool
    fee: Optional[uint64]

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> CancelOfferRequest:
        return cls(
            trade_id=bytes32.from_hexstr(marshalled["trade_id"]),
            secure=marshalled["secure"],
            fee=None if marshalled["fee"] is None else uint64(marshalled["fee"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "trade_id": self.trade_id.hex(),
            "secure": self.secure,
            "fee": None if self.fee is None else int(self.fee),
        }


@dataclasses.dataclass(frozen=True)
class CancelOfferResponse:
    success: bool

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> CancelOfferResponse:
        return cls(
            success=marshalled["success"],
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "success": self.success,
        }


@final
@dataclasses.dataclass(frozen=True)
class ClearPendingRootsRequest:
    store_id: bytes32

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> ClearPendingRootsRequest:
        return cls(
            store_id=bytes32.from_hexstr(marshalled["store_id"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "store_id": self.store_id.hex(),
        }


@final
@dataclasses.dataclass(frozen=True)
class ClearPendingRootsResponse:
    success: bool

    root: Optional[Root]
    # store_id: bytes32
    # node_hash: Optional[bytes32]
    # generation: int
    # status: Status

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> ClearPendingRootsResponse:
        return cls(
            success=marshalled["success"],
            root=None if marshalled["root"] is None else Root.unmarshal(marshalled["root"]),
        )

    def marshal(self) -> dict[str, Any]:
        return {
            "success": self.success,
            "root": None if self.root is None else self.root.marshal(),
        }


@dataclasses.dataclass(frozen=True)
class SyncStatus:
    root_hash: bytes32
    generation: int
    target_root_hash: bytes32
    target_generation: int


@final
@dataclasses.dataclass(frozen=True)
class PluginRemote:
    url: str
    # repr=False to avoid leaking secrets
    headers: dict[str, str] = dataclasses.field(default_factory=dict, hash=False, repr=False)

    @classmethod
    def unmarshal(cls, marshalled: dict[str, Any]) -> PluginRemote:
        return cls(
            url=marshalled["url"],
            headers=marshalled.get("headers", {}),
        )


@dataclasses.dataclass(frozen=True)
class PluginStatus:
    uploaders: dict[str, dict[str, Any]]
    downloaders: dict[str, dict[str, Any]]

    def marshal(self) -> dict[str, Any]:
        return {
            "plugin_status": {
                "uploaders": self.uploaders,
                "downloaders": self.downloaders,
            }
        }


@dataclasses.dataclass(frozen=True)
class InsertResult:
    node_hash: bytes32
    root: Root


@dataclasses.dataclass(frozen=True)
class UnsubscribeData:
    store_id: bytes32
    retain_data: bool


@dataclasses.dataclass(frozen=True)
class KeysValuesCompressed:
    keys_values_hashed: dict[bytes32, bytes32]
    key_hash_to_length: dict[bytes32, int]
    leaf_hash_to_length: dict[bytes32, int]
    root_hash: Optional[bytes32]


@dataclasses.dataclass(frozen=True)
class KeysPaginationData:
    total_pages: int
    total_bytes: int
    keys: list[bytes]
    root_hash: Optional[bytes32]


@dataclasses.dataclass(frozen=True)
class KeysValuesPaginationData:
    total_pages: int
    total_bytes: int
    keys_values: list[TerminalNode]
    root_hash: Optional[bytes32]


@dataclasses.dataclass(frozen=True)
class KVDiffPaginationData:
    total_pages: int
    total_bytes: int
    kv_diff: list[DiffData]


#
# GetProof and VerifyProof support classes
#
@streamable
@dataclasses.dataclass(frozen=True)
class ProofLayer(Streamable):
    # This class is basically Layer but streamable
    other_hash_side: uint8
    other_hash: bytes32
    combined_hash: bytes32


@streamable
@dataclasses.dataclass(frozen=True)
class HashOnlyProof(Streamable):
    key_klvm_hash: bytes32
    value_klvm_hash: bytes32
    node_hash: bytes32
    layers: list[ProofLayer]

    def root(self) -> bytes32:
        if len(self.layers) == 0:
            return self.node_hash
        return self.layers[-1].combined_hash

    @classmethod
    def from_key_value(cls, key: bytes, value: bytes, node_hash: bytes32, layers: list[ProofLayer]) -> HashOnlyProof:
        return cls(
            key_klvm_hash=Program.to(key).get_tree_hash(),
            value_klvm_hash=Program.to(value).get_tree_hash(),
            node_hash=node_hash,
            layers=layers,
        )


@streamable
@dataclasses.dataclass(frozen=True)
class KeyValueHashes(Streamable):
    key_klvm_hash: bytes32
    value_klvm_hash: bytes32


@streamable
@dataclasses.dataclass(frozen=True)
class ProofResultInclusions(Streamable):
    store_id: bytes32
    inclusions: list[KeyValueHashes]


@streamable
@dataclasses.dataclass(frozen=True)
class GetProofRequest(Streamable):
    store_id: bytes32
    keys: list[bytes]


@streamable
@dataclasses.dataclass(frozen=True)
class StoreProofsHashes(Streamable):
    store_id: bytes32
    proofs: list[HashOnlyProof]


@streamable
@dataclasses.dataclass(frozen=True)
class DLProof(Streamable):
    store_proofs: StoreProofsHashes
    coin_id: bytes32
    inner_puzzle_hash: bytes32


@streamable
@dataclasses.dataclass(frozen=True)
class GetProofResponse(Streamable):
    proof: DLProof
    success: bool


@streamable
@dataclasses.dataclass(frozen=True)
class VerifyProofResponse(Streamable):
    verified_klvm_hashes: ProofResultInclusions
    current_root: bool
    success: bool


def dl_verify_proof_internal(dl_proof: DLProof, puzzle_hash: bytes32) -> list[KeyValueHashes]:
    """Verify a proof of inclusion for a DL singleton"""

    verified_keys: list[KeyValueHashes] = []

    for reference_proof in dl_proof.store_proofs.proofs:
        inner_puz_hash = dl_proof.inner_puzzle_hash
        host_fullpuz_program = create_host_fullpuz(
            inner_puz_hash, reference_proof.root(), dl_proof.store_proofs.store_id
        )
        expected_puzzle_hash = host_fullpuz_program.get_tree_hash_precalc(inner_puz_hash)

        if puzzle_hash != expected_puzzle_hash:
            raise ProofIntegrityError(
                "Invalid Proof: incorrect puzzle hash: expected:"
                f"{expected_puzzle_hash.hex()} received: {puzzle_hash.hex()}"
            )

        proof = ProofOfInclusion(
            node_hash=reference_proof.node_hash,
            layers=[
                ProofOfInclusionLayer(
                    other_hash_side=Side(layer.other_hash_side),
                    other_hash=layer.other_hash,
                    combined_hash=layer.combined_hash,
                )
                for layer in reference_proof.layers
            ],
        )

        leaf_hash = internal_hash(left_hash=reference_proof.key_klvm_hash, right_hash=reference_proof.value_klvm_hash)
        if leaf_hash != proof.node_hash:
            raise ProofIntegrityError("Invalid Proof: node hash does not match key and value")

        if not proof.valid():
            raise ProofIntegrityError("Invalid Proof: invalid proof of inclusion found")

        verified_keys.append(
            KeyValueHashes(key_klvm_hash=reference_proof.key_klvm_hash, value_klvm_hash=reference_proof.value_klvm_hash)
        )

    return verified_keys


async def dl_verify_proof(
    request: dict[str, Any],
    wallet_node: WalletNode,
    peer: WSChikConnection,
) -> dict[str, Any]:
    """Verify a proof of inclusion for a DL singleton"""

    dlproof = DLProof.from_json_dict(request)

    coin_id = dlproof.coin_id
    coin_states = await wallet_node.get_coin_state([coin_id], peer=peer)
    if len(coin_states) == 0:
        raise ProofIntegrityError(f"Invalid Proof: No DL singleton found at coin id: {coin_id.hex()}")

    verified_keys = dl_verify_proof_internal(dlproof, coin_states[0].coin.puzzle_hash)

    response = VerifyProofResponse(
        verified_klvm_hashes=ProofResultInclusions(dlproof.store_proofs.store_id, verified_keys),
        success=True,
        current_root=coin_states[0].spent_height is None,
    )
    return response.to_json_dict()



from __future__ import annotations

import os
import pathlib
import sys
import time
from collections.abc import AsyncIterable, Awaitable, Iterator
from typing import Any, Callable

import pytest

# https://github.com/pytest-dev/pytest/issues/7469
from _pytest.fixtures import SubRequest

from chik._tests.core.data_layer.util import (
    ChikRoot,
    Example,
    add_0123_example,
    add_01234567_example,
    create_valid_node_values,
)
from chik._tests.util.misc import closing_chik_root_popen
from chik.data_layer.data_layer_util import NodeType, Status
from chik.data_layer.data_store import DataStore
from chik.types.blockchain_format.sized_bytes import bytes32

# TODO: These are more general than the data layer and should either move elsewhere or
#       be replaced with an existing common approach.  For now they can at least be
#       shared among the data layer test files.


@pytest.fixture(name="chik_daemon", scope="function")
def chik_daemon_fixture(chik_root: ChikRoot) -> Iterator[None]:
    with closing_chik_root_popen(chik_root=chik_root, args=[sys.executable, "-m", "chik.daemon.server"]):
        # TODO: this is not pretty as a hard coded time
        # let it settle
        time.sleep(5)
        yield


@pytest.fixture(name="chik_data", scope="function")
def chik_data_fixture(chik_root: ChikRoot, chik_daemon: None, scripts_path: pathlib.Path) -> Iterator[None]:
    with closing_chik_root_popen(chik_root=chik_root, args=[os.fspath(scripts_path.joinpath("chik_data_layer"))]):
        # TODO: this is not pretty as a hard coded time
        # let it settle
        time.sleep(5)
        yield


@pytest.fixture(name="create_example", params=[add_0123_example, add_01234567_example])
def create_example_fixture(request: SubRequest) -> Callable[[DataStore, bytes32], Awaitable[Example]]:
    # https://github.com/pytest-dev/pytest/issues/8763
    return request.param  # type: ignore[no-any-return]


@pytest.fixture(name="store_id", scope="function")
def store_id_fixture() -> bytes32:
    base = b"a store id"
    pad = b"." * (32 - len(base))
    return bytes32(pad + base)


@pytest.fixture(name="raw_data_store", scope="function")
async def raw_data_store_fixture(database_uri: str) -> AsyncIterable[DataStore]:
    async with DataStore.managed(database=database_uri, uri=True) as store:
        yield store


@pytest.fixture(name="data_store", scope="function")
async def data_store_fixture(raw_data_store: DataStore, store_id: bytes32) -> AsyncIterable[DataStore]:
    await raw_data_store.create_tree(store_id=store_id, status=Status.COMMITTED)

    await raw_data_store.check()
    yield raw_data_store
    await raw_data_store.check()


@pytest.fixture(name="node_type", params=NodeType)
def node_type_fixture(request: SubRequest) -> NodeType:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture(name="valid_node_values")
async def valid_node_values_fixture(
    data_store: DataStore,
    store_id: bytes32,
    node_type: NodeType,
) -> dict[str, Any]:
    await add_01234567_example(data_store=data_store, store_id=store_id)

    if node_type == NodeType.INTERNAL:
        node_a = await data_store.get_node_by_key(key=b"\x02", store_id=store_id)
        node_b = await data_store.get_node_by_key(key=b"\x04", store_id=store_id)
        return create_valid_node_values(node_type=node_type, left_hash=node_a.hash, right_hash=node_b.hash)
    elif node_type == NodeType.TERMINAL:
        return create_valid_node_values(node_type=node_type)
    else:
        raise Exception(f"invalid node type: {node_type!r}")


@pytest.fixture(name="bad_node_type", params=range(2 * len(NodeType)))
def bad_node_type_fixture(request: SubRequest, valid_node_values: dict[str, Any]) -> int:
    if request.param == valid_node_values["node_type"]:
        pytest.skip("Actually, this is a valid node type")

    return request.param  # type: ignore[no-any-return]



from __future__ import annotations

from chik_rs import G1Element, G2Element, PrivateKey

from chik.full_node.bundle_tools import simple_solution_generator
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import make_spend
from chik.types.condition_opcodes import ConditionOpcode
from chik.types.generator_types import BlockGenerator
from chik.types.spend_bundle import SpendBundle
from chik.util.ints import uint64
from chik.wallet.puzzles.p2_delegated_puzzle_or_hidden_puzzle import puzzle_for_pk, solution_for_conditions

GROUP_ORDER = 0x73EDA753299D7D483339D80809A1D80553BDA402FFFE5BFEFFFFFFFF00000001


def int_to_public_key(index: int) -> G1Element:
    index %= GROUP_ORDER
    private_key_from_int = PrivateKey.from_bytes(index.to_bytes(32, "big"))
    return private_key_from_int.get_g1()


def puzzle_hash_for_index(index: int, puzzle_hash_db: dict[bytes32, SerializedProgram]) -> bytes32:
    public_key: G1Element = int_to_public_key(index)
    puzzle = SerializedProgram.from_program(puzzle_for_pk(public_key))
    puzzle_hash: bytes32 = puzzle.get_tree_hash()
    puzzle_hash_db[puzzle_hash] = puzzle
    return puzzle_hash


def make_fake_coin(index: int, puzzle_hash_db: dict[bytes32, SerializedProgram]) -> Coin:
    """
    Make a fake coin with parent id equal to the index (ie. a genesis block coin)

    """
    parent: bytes32 = bytes32(index.to_bytes(32, "big"))
    puzzle_hash: bytes32 = puzzle_hash_for_index(index, puzzle_hash_db)
    amount: uint64 = uint64(100000)
    return Coin(parent, puzzle_hash, amount)


def conditions_for_payment(coin: Coin) -> Program:
    d: dict[bytes32, SerializedProgram] = {}  # a throwaway db since we don't care
    new_puzzle_hash = puzzle_hash_for_index(int.from_bytes(coin.puzzle_hash, "big"), d)
    ret: Program = Program.to([[ConditionOpcode.CREATE_COIN, new_puzzle_hash, coin.amount]])
    return ret


def make_spend_bundle(count: int) -> SpendBundle:
    puzzle_hash_db: dict[bytes32, SerializedProgram] = {}
    coins = [make_fake_coin(_, puzzle_hash_db) for _ in range(count)]

    coin_spends = []
    for coin in coins:
        puzzle_reveal = puzzle_hash_db[coin.puzzle_hash]
        conditions = conditions_for_payment(coin)
        solution = SerializedProgram.from_program(solution_for_conditions(conditions))
        coin_spend = make_spend(coin, puzzle_reveal, solution)
        coin_spends.append(coin_spend)

    spend_bundle = SpendBundle(coin_spends, G2Element())
    return spend_bundle


def make_block_generator(count: int) -> BlockGenerator:
    spend_bundle = make_spend_bundle(count)
    return simple_solution_generator(spend_bundle)



from __future__ import annotations

import asyncio
import json
import logging
import os
import sys
import traceback
from pathlib import Path
from typing import Any, Literal, Optional, Union

from chik.plotters.plotters_util import get_venv_bin, reset_loop_policy_for_windows, run_command, run_plotter
from chik.plotting.create_plots import resolve_plot_keys

log = logging.getLogger(__name__)


BLADEBIT_PLOTTER_DIR = "bladebit"


def is_bladebit_supported() -> bool:
    # bladebit >= 2.0.0 now supports macOS
    return sys.platform.startswith("linux") or sys.platform in {"win32", "cygwin", "darwin"}


def meets_memory_requirement(plotters_root_path: Path) -> tuple[bool, Optional[str]]:
    have_enough_memory: bool = False
    warning_string: Optional[str] = None

    bladebit_executable_path = get_bladebit_executable_path(plotters_root_path)
    if bladebit_executable_path.exists():
        try:
            proc = run_command(
                [os.fspath(bladebit_executable_path), "--memory-json"],
                "Failed to call bladebit with --memory-json option",
                capture_output=True,
                text=True,
                check=False,
            )
            if proc.returncode != 0:
                return have_enough_memory, proc.stderr.strip()

            memory_info: dict[str, int] = json.loads(proc.stdout)
            total_bytes: int = memory_info.get("total", -1)
            required_bytes: int = memory_info.get("required", 0)
            have_enough_memory = total_bytes >= required_bytes
            if have_enough_memory is False:
                warning_string = f"BladeBit requires at least {int(required_bytes / 1024**3)} GiB of RAM to operate"
        except Exception as e:
            print(f"Failed to determine bladebit memory requirements: {e}")

    return have_enough_memory, warning_string


def is_cudaplot_available(plotters_root_path: Path) -> bool:
    bladebit_executable_path = get_bladebit_executable_path(plotters_root_path)
    if not bladebit_executable_path.exists():
        return False
    try:
        proc = run_command(
            [os.fspath(bladebit_executable_path), "cudacheck"],
            "Failed to call bladebit with cudacheck command",
            capture_output=True,
            text=True,
            check=False,
        )
        return proc.returncode == 0
    except Exception as e:
        print(f"Failed to determine whether bladebit supports cuda: {e}")
        return False


def get_bladebit_src_path(plotters_root_path: Path) -> Path:
    return plotters_root_path / BLADEBIT_PLOTTER_DIR


def get_bladebit_package_path() -> Path:
    p = Path(os.path.dirname(sys.executable)).joinpath("_internal/bladebit")
    if p.exists():
        return p
    return Path(os.path.dirname(sys.executable)).joinpath("bladebit")


def get_bladebit_exec_path(with_cuda: bool = False) -> str:
    if with_cuda:
        return "bladebit_cuda.exe" if sys.platform in {"win32", "cygwin"} else "bladebit_cuda"
    return "bladebit.exe" if sys.platform in {"win32", "cygwin"} else "bladebit"


def get_bladebit_exec_venv_path(with_cuda: bool = False) -> Optional[Path]:
    venv_bin_path = get_venv_bin()
    if not venv_bin_path:
        return None
    bladebit_exec = get_bladebit_exec_path(with_cuda)
    return venv_bin_path / bladebit_exec


def get_bladebit_exec_src_path(plotters_root_path: Path, with_cuda: bool = False) -> Path:
    bladebit_src_dir = get_bladebit_src_path(plotters_root_path)
    build_dir = "build/Release" if sys.platform in {"win32", "cygwin"} else "build"
    bladebit_exec = get_bladebit_exec_path(with_cuda)
    return bladebit_src_dir / build_dir / bladebit_exec


def get_bladebit_exec_package_path(with_cuda: bool = False) -> Path:
    bladebit_package_dir = get_bladebit_package_path()
    bladebit_exec = get_bladebit_exec_path(with_cuda)
    return bladebit_package_dir / bladebit_exec


def get_bladebit_executable_path(plotters_root_path: Path) -> Path:
    # Search for bladebit executable which supports CUDA at the first priority
    bladebit_exec_venv_path = get_bladebit_exec_venv_path(with_cuda=True)
    if bladebit_exec_venv_path is not None and bladebit_exec_venv_path.exists():
        return bladebit_exec_venv_path
    bladebit_exec_src_path = get_bladebit_exec_src_path(plotters_root_path, with_cuda=True)
    if bladebit_exec_src_path.exists():
        return bladebit_exec_src_path
    bladebit_exec_package_path = get_bladebit_exec_package_path(with_cuda=True)
    if bladebit_exec_package_path.exists():
        return bladebit_exec_package_path

    bladebit_exec_venv_path = get_bladebit_exec_venv_path(with_cuda=False)
    if bladebit_exec_venv_path is not None and bladebit_exec_venv_path.exists():
        return bladebit_exec_venv_path
    bladebit_exec_src_path = get_bladebit_exec_src_path(plotters_root_path, with_cuda=False)
    if bladebit_exec_src_path.exists():
        return bladebit_exec_src_path
    return get_bladebit_exec_package_path(with_cuda=False)


def get_bladebit_version(
    plotters_root_path: Path,
) -> Union[tuple[Literal[False], str], tuple[None, str], tuple[Literal[True], list[str]]]:
    bladebit_executable_path = get_bladebit_executable_path(plotters_root_path)
    if not bladebit_executable_path.exists():
        # (found=False, "")
        return False, ""

    try:
        proc = run_command(
            [os.fspath(bladebit_executable_path), "--version"],
            "Failed to call bladebit with --version option",
            capture_output=True,
            text=True,
            check=False,
        )
        if proc.returncode != 0:
            # (found=unknown, errMsg)
            return None, proc.stderr.strip()

        # (found=True, versionStr)
        version_str: str = proc.stdout.strip()
        return True, version_str.split(".")
    except Exception as e:
        # (found=unknown, errMsg)
        return None, str(e)


def get_bladebit_install_info(plotters_root_path: Path) -> Optional[dict[str, Any]]:
    info: dict[str, Any] = {"display_name": "BladeBit Plotter"}
    installed: bool = False
    supported: bool = is_bladebit_supported()
    cuda_available: bool = is_cudaplot_available(plotters_root_path)

    bladebit_executable_path = get_bladebit_executable_path(plotters_root_path)
    if bladebit_executable_path.exists():
        version: Optional[str] = None
        found, response = get_bladebit_version(plotters_root_path)
        if found:
            version = ".".join(response)
        elif found is None:
            print(f"Failed to determine bladebit version: {response}")

        if version is not None:
            installed = True
            info["version"] = version
        else:
            installed = False

    info["installed"] = installed
    if installed is False:
        info["can_install"] = supported

    if supported:
        _, memory_warning = meets_memory_requirement(plotters_root_path)
        if memory_warning is not None:
            info["bladebit_memory_warning"] = memory_warning

    info["cuda_support"] = cuda_available

    return info


progress_bladebit_cuda = {
    "Generating F1": 0.01,
    "Finished F1 in ": 0.1,
    "Table 2 completed in ": 0.2,
    "Table 3 completed in ": 0.3,
    "Table 4 completed in ": 0.4,
    "Table 5 completed in ": 0.5,
    "Table 6 completed in ": 0.6,
    "Table 7 completed in ": 0.7,
    "Completed Phase 1 in ": 0.8,
    "Completed Phase 2 in ": 0.9,
    "Completed Phase 3 in ": 0.95,
}

progress_bladebit_ram = {
    "Finished F1 sort": 0.01,
    "Finished forward propagating table 2": 0.06,
    "Finished forward propagating table 3": 0.12,
    "Finished forward propagating table 4": 0.2,
    "Finished forward propagating table 5": 0.28,
    "Finished forward propagating table 6": 0.36,
    "Finished forward propagating table 7": 0.42,
    "Finished prunning table 6": 0.43,
    "Finished prunning table 5": 0.48,
    "Finished prunning table 4": 0.51,
    "Finished prunning table 3": 0.55,
    "Finished prunning table 2": 0.58,
    "Finished compressing tables 1 and 2": 0.66,
    "Finished compressing tables 2 and 3": 0.73,
    "Finished compressing tables 3 and 4": 0.79,
    "Finished compressing tables 4 and 5": 0.85,
    "Finished compressing tables 5 and 6": 0.92,
    "Finished compressing tables 6 and 7": 0.98,
}

progress_bladebit_disk = {
    # "Running Phase 1": 0.01,
    "Finished f1 generation in ": 0.01,
    "Completed table 2 in ": 0.06,
    "Completed table 3 in ": 0.12,
    "Completed table 4 in ": 0.2,
    "Completed table 5 in ": 0.28,
    "Completed table 6 in ": 0.36,
    "Completed table 7 in ": 0.42,
    # "Finished Phase 1 ": 0.43,
    # "Running Phase 2": 0.43,
    "Finished marking table 6 in ": 0.43,
    "Finished marking table 5 in ": 0.48,
    "Finished marking table 4 in ": 0.51,
    "Finished marking table 3 in ": 0.55,
    "Finished marking table 2 in ": 0.58,
    # "Finished Phase 2 ": 0.59,
    # "Running Phase 3": 0.60,
    "Finished compressing tables 1 and 2 in ": 0.66,
    "Finished compressing tables 2 and 3 in ": 0.73,
    "Finished compressing tables 3 and 4 in ": 0.79,
    "Finished compressing tables 4 and 5 in ": 0.85,
    "Finished compressing tables 5 and 6 in ": 0.92,
    "Finished compressing tables 6 and 7 in ": 0.98,
    # "Finished Phase 3 ": 0.99,
    "Finished writing plot ": 0.99,
}


def plot_bladebit(args, chik_root_path, root_path):
    (found, version_or_exception) = get_bladebit_version(root_path)
    if found is None:
        print(f"Error: {version_or_exception}")
        return

    if found and int(version_or_exception[0]) < 2:
        print(f"Version {'.'.join(version_or_exception)} is detected.")
        print("bladebit < 2 is not supported any more.")
        return

    bladebit_executable_path = get_bladebit_executable_path(root_path)
    if not os.path.exists(bladebit_executable_path):
        print("Bladebit was not found.")
        return

    if sys.platform in {"win32", "cygwin"}:
        reset_loop_policy_for_windows()

    plot_keys = asyncio.run(
        resolve_plot_keys(
            None if args.farmerkey == b"" else args.farmerkey.hex(),
            None,
            None if args.pool_key == b"" else args.pool_key.hex(),
            None if args.contract == "" else args.contract,
            chik_root_path,
            log,
            args.connect_to_daemon,
        )
    )
    if args.plot_type in {"ramplot", "diskplot", "cudaplot"}:
        plot_type = args.plot_type
    else:
        plot_type = "diskplot"
        print("plot_type is automatically set to diskplot")

    call_args = [
        os.fspath(bladebit_executable_path),
        "--threads",
        str(args.threads),
        "--count",
        str(args.count),
        "--farmer-key",
        bytes(plot_keys.farmer_public_key).hex(),
    ]
    if plot_keys.pool_public_key is not None:
        call_args.append("--pool-key")
        call_args.append(bytes(plot_keys.pool_public_key).hex())
    if plot_keys.pool_contract_address is not None:
        call_args.append("--pool-contract")
        call_args.append(plot_keys.pool_contract_address)
    if args.warmstart:
        call_args.append("--warm-start")
    if args.id is not None and args.id != b"":
        call_args.append("--plot-id")
        call_args.append(args.id.hex())
    if "memo" in args and args.memo is not None and args.memo != b"":
        call_args.append("--memo")
        call_args.append(args.memo)
    if args.nonuma:
        call_args.append("--no-numa")
    if args.no_cpu_affinity:
        call_args.append("--no-cpu-affinity")
    if args.verbose:
        call_args.append("--verbose")
    if (
        "compress" in args
        and args.compress is not None
        and str(args.compress).isdigit()
        and int(version_or_exception[0]) >= 3
    ):
        call_args.append("--compress")
        call_args.append(str(args.compress))

    call_args.append(plot_type)

    if "buckets" in args and args.buckets:
        call_args.append("--buckets")
        call_args.append(str(args.buckets))
    if "tmpdir" in args and args.tmpdir:
        call_args.append("--temp1")
        call_args.append(str(args.tmpdir))
    if "tmpdir2" in args and args.tmpdir2:
        call_args.append("--temp2")
        call_args.append(str(args.tmpdir2))
    if "cache" in args and args.cache is not None:
        call_args.append("--cache")
        call_args.append(str(args.cache))
    if "f1_threads" in args and args.f1_threads:
        call_args.append("--f1-threads")
        call_args.append(str(args.f1_threads))
    if "fp_threads" in args and args.fp_threads:
        call_args.append("--fp-threads")
        call_args.append(str(args.fp_threads))
    if "c_threads" in args and args.c_threads:
        call_args.append("--c-threads")
        call_args.append(str(args.c_threads))
    if "p2_threads" in args and args.p2_threads:
        call_args.append("--p2-threads")
        call_args.append(str(args.p2_threads))
    if "p3_threads" in args and args.p3_threads:
        call_args.append("--p3-threads")
        call_args.append(str(args.p3_threads))
    if "alternate" in args and args.alternate:
        call_args.append("--alternate")
    if "no_t1_direct" in args and args.no_t1_direct:
        call_args.append("--no-t1-direct")
    if "no_t2_direct" in args and args.no_t2_direct:
        call_args.append("--no-t2-direct")
    if "device" in args and str(args.device).isdigit():
        call_args.append("--device")
        call_args.append(str(args.device))
    if "disk_128" in args and args.disk_128:
        call_args.append("--disk-128")
    if "disk_16" in args and args.disk_16:
        call_args.append("--disk-16")

    call_args.append(args.finaldir)

    try:
        if plot_type == "cudaplot":
            progress = progress_bladebit_cuda
        elif plot_type == "ramplot":
            progress = progress_bladebit_ram
        else:
            progress = progress_bladebit_disk
        asyncio.run(run_plotter(chik_root_path, args.plotter, call_args, progress))
    except Exception as e:
        print(f"Exception while plotting: {e} {type(e)}")
        print(f"Traceback: {traceback.format_exc()}")
